"""
An action is a way to take an arbitrary method and make it configurable and runnable within a Data Context.

The only requirement from an action is for it to have a take_action method.
"""  # noqa: E501 # FIXME CoP

from __future__ import annotations

import json
import logging
import smtplib
import ssl
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from typing import (
    TYPE_CHECKING,
    Any,
    List,
    Literal,
    Optional,
    Type,
    Union,
)

import requests
from typing_extensions import dataclass_transform

from great_expectations._docs_decorators import public_api
from great_expectations.analytics.client import submit as submit_event
from great_expectations.analytics.events import (
    NotificationActionRanEvent,
)
from great_expectations.compatibility import aws
from great_expectations.compatibility.pydantic import (
    BaseModel,
    Extra,
    Field,
    ModelMetaclass,
    root_validator,
    validator,
)
from great_expectations.compatibility.pypd import pypd
from great_expectations.compatibility.typing_extensions import override
from great_expectations.data_context.cloud_constants import GXCloudRESTResource
from great_expectations.data_context.data_context.context_factory import project_manager
from great_expectations.data_context.types.resource_identifiers import (
    ExpectationSuiteIdentifier,
    GXCloudIdentifier,
    ValidationResultIdentifier,
)
from great_expectations.data_context.util import instantiate_class_from_config
from great_expectations.datasource.fluent.config_str import ConfigStr
from great_expectations.exceptions import ClassInstantiationError
from great_expectations.exceptions.exceptions import (
    ValidationActionAlreadyRegisteredError,
    ValidationActionRegistryRetrievalError,
)
from great_expectations.render.renderer import (
    EmailRenderer,
    MicrosoftTeamsRenderer,
    OpsgenieRenderer,
    SlackRenderer,
)
from great_expectations.render.renderer.renderer import Renderer
from great_expectations.util import convert_to_json_serializable  # noqa: TID251 # FIXME CoP

if TYPE_CHECKING:
    from great_expectations.checkpoint.checkpoint import CheckpointResult
    from great_expectations.core.expectation_validation_result import (
        ExpectationSuiteValidationResult,
    )

logger = logging.getLogger(__name__)


def _build_renderer(config: dict) -> Renderer:
    renderer = instantiate_class_from_config(
        config=config,
        runtime_environment={},
        config_defaults={"module_name": "great_expectations.render.renderer"},
    )
    if not renderer:
        raise ClassInstantiationError(
            module_name=config.get("module_name"),
            package_name=None,
            class_name=config.get("class_name"),
        )
    return renderer


@public_api
class ActionContext:
    """
    Shared context for all Actions in a Checkpoint run.
    Note that order matters in the Action list, as the context is updated with each Action's result.
    """

    def __init__(self) -> None:
        self._data: list[tuple[ValidationAction, dict]] = []

    @property
    def data(self) -> list[tuple[ValidationAction, dict]]:
        return self._data

    def update(self, action: ValidationAction, action_result: dict) -> None:
        self._data.append((action, action_result))

    @public_api
    def filter_results(self, class_: Type[ValidationAction]) -> list[dict]:
        """
        Filter the results of the actions in the context by class.

        Args:
            class_: The class to filter by.

        Returns:
            A list of action results.
        """
        return [action_result for action, action_result in self._data if isinstance(action, class_)]


class ValidationActionRegistry:
    """
    Registers ValidationActions to enable deserialization based on their configuration.

    Uses the `type` key from the action configuration to determine which registered class
    to instantiate.
    """

    def __init__(self):
        self._registered_actions: dict[str, Type[ValidationAction]] = {}

    def register(self, action_type: str, action_class: Type[ValidationAction]) -> None:
        """
        Register a ValidationAction class with the registry.

        Args:
            action_type: The type of the action to register.
            action_class: The ValidationAction class to register.

        Raises:
            ValidationActionAlreadyRegisteredError: If the action type is already registered.
        """
        if action_type in self._registered_actions:
            raise ValidationActionAlreadyRegisteredError(action_type)

        self._registered_actions[action_type] = action_class

    def get(self, action_type: str | None) -> Type[ValidationAction]:
        """
        Return a ValidationAction class based on its type.
        Used when instantiating actions from a checkpoint configuration.

        Args:
            action_type: The 'type' key from the action configuration.

        Returns:
            The ValidationAction class corresponding to the configuration.

        Raises:
            ValidationActionRegistryRetrievalError: If the action type is not registered.
        """
        if action_type not in self._registered_actions:
            raise ValidationActionRegistryRetrievalError(action_type)

        return self._registered_actions[action_type]


_VALIDATION_ACTION_REGISTRY = ValidationActionRegistry()


@dataclass_transform(kw_only_default=True, field_specifiers=(Field,))  # Enables type hinting
class MetaValidationAction(ModelMetaclass):
    """MetaValidationAction registers ValidationAction as they are defined, adding them to
    the registry.

    Any class inheriting from ValidationAction will be registered based on the value of the
    "type" class attribute.
    """

    def __new__(cls, clsname, bases, attrs):
        newclass = super().__new__(cls, clsname, bases, attrs)

        action_type = newclass.__fields__.get("type")
        if action_type and action_type.default:  # Excludes base classes
            _VALIDATION_ACTION_REGISTRY.register(
                action_type=action_type.default, action_class=newclass
            )

        return newclass


@public_api
class ValidationAction(BaseModel, metaclass=MetaValidationAction):
    """
    Actions define a set of steps to run after a Validation Result is produced. Subclass `ValidationAction` to create a [custom Action](/docs/core/trigger_actions_based_on_results/create_a_custom_action).

    Through a Checkpoint, one can orchestrate the validation of data and configure notifications, data documentation updates,
    and other actions to take place after the Validation Result is produced.
    """  # noqa: E501 # FIXME CoP

    class Config:
        extra = Extra.forbid
        arbitrary_types_allowed = True
        # Due to legacy pattern of instantiate_class_from_config, we need a custom serializer
        json_encoders = {Renderer: lambda r: r.serialize()}

    type: str
    name: str

    @property
    def _using_cloud_context(self) -> bool:
        return project_manager.is_using_cloud()

    @public_api
    def run(
        self, checkpoint_result: CheckpointResult, action_context: ActionContext | None = None
    ) -> dict:
        """
        Run the action.

        Args:
            checkpoint_result: The result of the checkpoint run.
            action_context: The context in which the action is run.

        Returns:
            A dictionary containing the result of the action.
        """
        raise NotImplementedError

    def _get_data_docs_pages_from_prior_action(
        self, action_context: ActionContext | None
    ) -> dict[ValidationResultIdentifier, dict[str, str]] | None:
        if action_context:
            data_docs_results = action_context.filter_results(class_=UpdateDataDocsAction)
            data_docs_pages = {}
            for result in data_docs_results:
                data_docs_pages.update(result)
            return data_docs_pages

        return None

    @staticmethod
    def _substitute_config_str_if_needed(value: Union[str, ConfigStr, None]) -> Optional[str]:
        from great_expectations.data_context.data_context.context_factory import project_manager

        config_provider = project_manager.get_config_provider()
        if isinstance(value, ConfigStr):
            return value.get_config_value(config_provider=config_provider)
        else:
            return value


def _should_notify(success: bool, notify_on: Literal["all", "failure", "success"]) -> bool:
    return (
        notify_on == "all"
        or (notify_on == "success" and success)
        or (notify_on == "failure" and not success)
    )


class DataDocsAction(ValidationAction):
    def _build_data_docs(
        self,
        site_names: list[str] | None = None,
        resource_identifiers: list | None = None,
    ) -> dict:
        return project_manager.build_data_docs(
            site_names=site_names, resource_identifiers=resource_identifiers
        )

    def _get_docs_sites_urls(
        self,
        site_names: list[str] | None = None,
        resource_identifier: Any | None = None,
    ):
        return project_manager.get_docs_sites_urls(
            site_names=site_names, resource_identifier=resource_identifier
        )


@public_api
class SlackNotificationAction(DataDocsAction):
    """Sends a Slack notification to a given webhook.

    ```yaml
    - name: send_slack_notification_on_validation_result
    action:
      class_name: SlackNotificationAction
      # put the actual webhook URL in the uncommitted/config_variables.yml file
      # or pass in as environment variable
      # use slack_webhook when not using slack bot token
      slack_webhook: ${validation_notification_slack_webhook}
      slack_token:
      slack_channel:
      notify_on: all
      notify_with:
      renderer:
        # the class that implements the message to be sent
        # this is the default implementation, but you can
        # implement a custom one
        module_name: great_expectations.render.renderer.slack_renderer
        class_name: SlackRenderer
      show_failed_expectations: True
    ```

    Args:
        renderer: Specifies the Renderer used to generate a query consumable by Slack API.
        slack_webhook: The incoming Slack webhook to which to send notification.
        slack_token: Token from Slack app. Used when not using slack_webhook.
        slack_channel: Slack channel to receive notification. Used when not using slack_webhook.
        notify_on: Specifies validation status that triggers notification. One of "all", "failure", "success".
        notify_with: List of DataDocs site names to display  in Slack messages. Defaults to all.
        show_failed_expectations: Shows a list of failed expectation types.

    Examples:
        **renderer:**

            ```python
            {
               "module_name": "great_expectations.render.renderer.slack_renderer",
               "class_name": "SlackRenderer",
           }
           ```
    """  # noqa: E501 # FIXME CoP

    type: Literal["slack"] = "slack"

    slack_webhook: Optional[Union[ConfigStr, str]] = None
    slack_token: Optional[Union[ConfigStr, str]] = None
    slack_channel: Optional[Union[ConfigStr, str]] = None
    notify_on: Literal["all", "failure", "success"] = "all"
    notify_with: Optional[List[str]] = None
    show_failed_expectations: bool = False
    renderer: SlackRenderer = Field(default_factory=SlackRenderer)

    @validator("renderer", pre=True)
    def _validate_renderer(cls, renderer: dict | SlackRenderer) -> SlackRenderer:
        if isinstance(renderer, dict):
            _renderer = _build_renderer(config=renderer)
            if not isinstance(_renderer, SlackRenderer):
                raise ValueError(  # noqa: TRY003, TRY004 # FIXME CoP
                    "renderer must be a SlackRenderer or a valid configuration for one."
                )
            renderer = _renderer
        return renderer

    @root_validator
    def _root_validate_slack_params(cls, values: dict) -> dict:
        slack_webhook = values["slack_webhook"]
        slack_token = values["slack_token"]
        slack_channel = values["slack_channel"]
        try:
            if slack_webhook:
                assert not slack_token and not slack_channel
            else:
                assert slack_token and slack_channel
        except AssertionError:
            raise ValueError("Please provide either slack_webhook or slack_token and slack_channel")  # noqa: TRY003 # FIXME CoP

        return values

    @override
    def run(
        self, checkpoint_result: CheckpointResult, action_context: ActionContext | None = None
    ) -> dict:
        success = checkpoint_result.success or False
        checkpoint_name = checkpoint_result.checkpoint_config.name
        result = {"slack_notification_result": "none required"}

        if not _should_notify(success=success, notify_on=self.notify_on):
            return result

        checkpoint_text_blocks: list[dict] = []
        for (
            validation_result_suite_identifier,
            validation_result_suite,
        ) in checkpoint_result.run_results.items():
            validation_text_blocks = self._render_validation_result(
                result_identifier=validation_result_suite_identifier,
                result=validation_result_suite,
                action_context=action_context,
            )
            checkpoint_text_blocks.extend(validation_text_blocks)

        payload = self.renderer.concatenate_text_blocks(
            action_name=self.name,
            text_blocks=checkpoint_text_blocks,
            success=success,
            checkpoint_name=checkpoint_name,
            run_id=checkpoint_result.run_id,
        )

        result = self._send_slack_notification(payload=payload)

        checkpoint = checkpoint_result.checkpoint_config
        submit_event(
            event=NotificationActionRanEvent(
                type=self.type, notify_type=self.notify_on, checkpoint_id=checkpoint.id
            )
        )

        return result

    def _render_validation_result(
        self,
        result_identifier: ValidationResultIdentifier,
        result: ExpectationSuiteValidationResult,
        action_context: ActionContext | None = None,
    ) -> list[dict]:
        data_docs_pages = None
        if action_context:
            data_docs_pages = self._get_data_docs_pages_from_prior_action(
                action_context=action_context
            )

        # Assemble complete GX Cloud URL for a specific validation result
        data_docs_urls: list[dict[str, str]] = self._get_docs_sites_urls(
            resource_identifier=result_identifier
        )

        validation_result_urls: list[str] = [
            data_docs_url["site_url"]
            for data_docs_url in data_docs_urls
            if data_docs_url["site_url"]
        ]
        if result.result_url:
            result.result_url += "?slack=true"
            validation_result_urls.append(result.result_url)

        return self.renderer.render(
            validation_result=result,
            data_docs_pages=data_docs_pages,
            notify_with=self.notify_with,
            validation_result_urls=validation_result_urls,
        )

    def _send_slack_notification(self, payload: dict) -> dict:
        slack_webhook = self._substitute_config_str_if_needed(self.slack_webhook)
        slack_token = self._substitute_config_str_if_needed(self.slack_token)
        slack_channel = self._substitute_config_str_if_needed(self.slack_channel)

        session = requests.Session()
        url = slack_webhook
        headers = None

        # Slack doc about overwritting the channel when using the legacy Incoming Webhooks
        # https://api.slack.com/legacy/custom-integrations/messaging/webhooks
        # ** Since it is legacy, it could be deprecated or removed in the future **
        if slack_channel:
            payload["channel"] = slack_channel

        if not slack_webhook:
            url = "https://slack.com/api/chat.postMessage"
            headers = {"Authorization": f"Bearer {slack_token}"}

        if not url:
            raise ValueError("No Slack webhook URL provided.")  # noqa: TRY003 # FIXME CoP

        try:
            response = session.post(url=url, headers=headers, json=payload)
            response.raise_for_status()
        except requests.ConnectionError:
            logger.warning(f"Failed to connect to Slack webhook after {10} retries.")
            return {"slack_notification_result": None}
        except requests.HTTPError:
            logger.warning(
                f"Request to Slack webhook returned error {response.status_code}: {response.text}"  # type: ignore[possibly-undefined] # ok for httperror
            )
            return {"slack_notification_result": None}

        return {"slack_notification_result": "Slack notification succeeded."}


class PagerdutyAlertAction(ValidationAction):
    """Sends a PagerDuty event.

    ```yaml
    - name: send_pagerduty_alert_on_validation_result
    action:
      class_name: PagerdutyAlertAction
      api_key: ${pagerduty_api_key}
      routing_key: ${pagerduty_routing_key}
      notify_on: failure
      severity: critical
    ```

    Args:
        api_key: Events API v2 key for pagerduty.
        routing_key: The 32 character Integration Key for an integration on a service or on a global ruleset.
        notify_on: Specifies validation status that triggers notification. One of "all", "failure", "success".
        severity: The PagerDuty severity levels determine the level of urgency. One of "critical", "error", "warning", or "info".
    """  # noqa: E501 # FIXME CoP

    type: Literal["pagerduty"] = "pagerduty"

    api_key: str
    routing_key: str
    notify_on: Literal["all", "failure", "success"] = "failure"
    severity: Literal["critical", "error", "warning", "info"] = "critical"

    @override
    def run(
        self, checkpoint_result: CheckpointResult, action_context: ActionContext | None = None
    ) -> dict:
        success = checkpoint_result.success or False
        checkpoint_name = checkpoint_result.checkpoint_config.name
        summary = f"Great Expectations Checkpoint {checkpoint_name} has "
        if success:
            summary += "succeeded"
        else:
            summary += "failed"

        return self._run_pypd_alert(dedup_key=checkpoint_name, message=summary, success=success)

    def _run_pypd_alert(self, dedup_key: str, message: str, success: bool):
        if _should_notify(success=success, notify_on=self.notify_on):
            pypd.api_key = self.api_key
            pypd.EventV2.create(
                data={
                    "routing_key": self.routing_key,
                    "dedup_key": dedup_key,
                    "event_action": "trigger",
                    "payload": {
                        "summary": message,
                        "severity": self.severity,
                        "source": "Great Expectations",
                    },
                }
            )

            return {"pagerduty_alert_result": "success"}

        return {"pagerduty_alert_result": "none sent"}


@public_api
class MicrosoftTeamsNotificationAction(ValidationAction):
    """Sends a Microsoft Teams notification to a given webhook.

    Args:
        teams_webhook: Incoming Microsoft Teams webhook to which to send notifications.
        notify_on: Specifies validation status that triggers notification. One of "all", "failure", "success".
    """  # noqa: E501 # FIXME CoP

    type: Literal["microsoft"] = "microsoft"

    teams_webhook: Union[ConfigStr, str]
    notify_on: Literal["all", "failure", "success"] = "all"
    renderer: MicrosoftTeamsRenderer = Field(default_factory=MicrosoftTeamsRenderer)

    @validator("renderer", pre=True)
    def _validate_renderer(cls, renderer: dict | MicrosoftTeamsRenderer) -> MicrosoftTeamsRenderer:
        if isinstance(renderer, dict):
            _renderer = _build_renderer(config=renderer)
            if not isinstance(_renderer, MicrosoftTeamsRenderer):
                raise ValueError(  # noqa: TRY003, TRY004 # FIXME CoP
                    "renderer must be a MicrosoftTeamsRenderer or a valid configuration for one."
                )
            renderer = _renderer
        return renderer

    @override
    def run(self, checkpoint_result: CheckpointResult, action_context: ActionContext | None = None):
        success = checkpoint_result.success or False
        if not _should_notify(success=success, notify_on=self.notify_on):
            return {"microsoft_teams_notification_result": None}

        data_docs_pages = self._get_data_docs_pages_from_prior_action(action_context=action_context)

        payload = self.renderer.render(
            checkpoint_result=checkpoint_result,
            data_docs_pages=data_docs_pages,
        )

        # this will actually sent the POST request to the Microsoft Teams webapp server
        teams_notif_result = self._send_microsoft_teams_notifications(payload=payload)

        checkpoint = checkpoint_result.checkpoint_config
        submit_event(
            event=NotificationActionRanEvent(
                type=self.type, notify_type=self.notify_on, checkpoint_id=checkpoint.id
            )
        )

        return {"microsoft_teams_notification_result": teams_notif_result}

    def _send_microsoft_teams_notifications(self, payload: dict) -> str | None:
        webhook = self._substitute_config_str_if_needed(self.teams_webhook)
        if not webhook:  # Necessary to appease mypy; this is guaranteed.
            raise ValueError("No Microsoft Teams webhook URL provided.")  # noqa: TRY003 # FIXME CoP

        session = requests.Session()
        try:
            response = session.post(url=webhook, json=payload)
            response.raise_for_status()
        except requests.ConnectionError:
            logger.warning("Failed to connect to Microsoft Teams webhook after 10 retries.")
            return None
        except requests.HTTPError as e:
            logger.warning(
                f"Request to Microsoft Teams API returned error {response.status_code}: {e}"  # type: ignore[possibly-undefined] # ok for httperror
            )
            return None

        return "Microsoft Teams notification succeeded."


class OpsgenieAlertAction(ValidationAction):
    """Sends an Opsgenie alert.

    ```yaml
    - name: send_opsgenie_alert_on_validation_result
    action:
      class_name: OpsgenieAlertAction
      # put the actual webhook URL in the uncommitted/config_variables.yml file
      # or pass in as environment variable
      api_key: ${opsgenie_api_key}
      region:
      priority: P2
      notify_on: failure
    ```

    Args:
        api_key: Opsgenie API key.
        region: Specifies the Opsgenie region. Populate 'EU' for Europe otherwise do not set.
        priority: Specifies the priority of the alert (P1 - P5).
        notify_on: Specifies validation status that triggers notification. One of "all", "failure", "success".
        tags: Tags to include in the alert
    """  # noqa: E501 # FIXME CoP

    type: Literal["opsgenie"] = "opsgenie"

    api_key: str
    region: Optional[str] = None
    priority: Literal["P1", "P2", "P3", "P4", "P5"] = "P3"
    notify_on: Literal["all", "failure", "success"] = "failure"
    tags: Optional[List[str]] = None
    renderer: OpsgenieRenderer = Field(default_factory=OpsgenieRenderer)

    @validator("renderer", pre=True)
    def _validate_renderer(cls, renderer: dict | OpsgenieRenderer) -> OpsgenieRenderer:
        if isinstance(renderer, dict):
            _renderer = _build_renderer(config=renderer)
            if not isinstance(_renderer, OpsgenieRenderer):
                raise ValueError(  # noqa: TRY003, TRY004 # FIXME CoP
                    "renderer must be a OpsgenieRenderer or a valid configuration for one."
                )
            renderer = _renderer
        return renderer

    @override
    def run(
        self, checkpoint_result: CheckpointResult, action_context: ActionContext | None = None
    ) -> dict:
        validation_success = checkpoint_result.success or False
        checkpoint_name = checkpoint_result.checkpoint_config.name

        if _should_notify(success=validation_success, notify_on=self.notify_on):
            settings = {
                "api_key": self.api_key,
                "region": self.region,
                "priority": self.priority,
                "tags": self.tags,
            }

            description = self.renderer.render(checkpoint_result=checkpoint_result)

            message = f"Great Expectations Checkpoint {checkpoint_name} "
            if checkpoint_result.success:
                message += "succeeded!"
            else:
                message += "failed!"

            alert_result = self._send_opsgenie_alert(
                query=description, message=message, settings=settings
            )

            return {"opsgenie_alert_result": alert_result}
        else:
            return {"opsgenie_alert_result": "No alert sent"}

    def _send_opsgenie_alert(self, query: str, message: str, settings: dict) -> bool:
        """Creates an alert in Opsgenie."""
        if settings["region"] is not None:
            # accommodate for Europeans
            url = f"https://api.{settings['region']}.opsgenie.com/v2/alerts"
        else:
            url = "https://api.opsgenie.com/v2/alerts"

        headers = {"Authorization": f"GenieKey {settings['api_key']}"}
        payload = {
            "message": message,
            "description": query,
            "priority": settings["priority"],  # allow this to be modified in settings
            "tags": settings["tags"],
        }

        session = requests.Session()

        try:
            response = session.post(url, headers=headers, json=payload)
            response.raise_for_status()
        except requests.ConnectionError as e:
            logger.warning(f"Failed to connect to Opsgenie: {e}")
            return False
        except requests.HTTPError as e:
            logger.warning(f"Request to Opsgenie API returned error {response.status_code}: {e}")  # type: ignore[possibly-undefined] # ok for httperror
            return False
        return True


@public_api
class EmailAction(ValidationAction):
    """Sends an email to a given list of email addresses.

    ```yaml
    - name: send_email_on_validation_result
    action:
      class_name: EmailAction
      notify_on: all # possible values: "all", "failure", "success"
      notify_with:
      renderer:
        # the class that implements the message to be sent
        # this is the default implementation, but you can
        # implement a custom one
        module_name: great_expectations.render.renderer.email_renderer
        class_name: EmailRenderer
      # put the actual following information in the uncommitted/config_variables.yml file
      # or pass in as environment variable
      smtp_address: ${smtp_address}
      smtp_port: ${smtp_port}
      sender_login: ${email_address}
      sender_password: ${sender_password}
      sender_alias: ${sender_alias} # useful to send an email as an alias
      receiver_emails: ${receiver_emails}
      use_tls: False
      use_ssl: True
    ```

    Args:
        renderer: Specifies the renderer used to generate an email.
        smtp_address: Address of the SMTP server used to send the email.
        smtp_address: Port of the SMTP server used to send the email.
        sender_login: Login used send the email.
        sender_password: Password used to send the email.
        sender_alias: Optional. Alias used to send the email (default = sender_login).
        receiver_emails: Email addresses that will receive the email (separated by commas).
        use_tls: Optional. Use of TLS to send the email (using either TLS or SSL is highly recommended).
        use_ssl: Optional. Use of SSL to send the email (using either TLS or SSL is highly recommended).
        notify_on: "Specifies validation status that triggers notification. One of "all", "failure", "success".
        notify_with: Optional list of DataDocs site names to display  in Slack messages. Defaults to all.

    Examples:
        **renderer:**

        ```python
        {
           "module_name": "great_expectations.render.renderer.email_renderer",
           "class_name": "EmailRenderer",
        }
        ```
    """  # noqa: E501 # FIXME CoP

    type: Literal["email"] = "email"

    smtp_address: Union[ConfigStr, str]
    smtp_port: Union[ConfigStr, str]
    receiver_emails: Union[ConfigStr, str]
    sender_login: Optional[Union[ConfigStr, str]] = None
    sender_password: Optional[Union[ConfigStr, str]] = None
    sender_alias: Optional[Union[ConfigStr, str]] = None
    use_tls: Optional[bool] = None
    use_ssl: Optional[bool] = None
    notify_on: Literal["all", "failure", "success"] = "all"
    notify_with: Optional[List[str]] = None
    renderer: EmailRenderer = Field(default_factory=EmailRenderer)

    @validator("renderer", pre=True)
    def _validate_renderer(cls, renderer: dict | EmailRenderer) -> EmailRenderer:
        if isinstance(renderer, dict):
            _renderer = _build_renderer(config=renderer)
            if not isinstance(_renderer, EmailRenderer):
                raise ValueError(  # noqa: TRY003, TRY004 # FIXME CoP
                    "renderer must be a EmailRenderer or a valid configuration for one."
                )
            renderer = _renderer
        return renderer

    @root_validator
    def _root_validate_email_params(cls, values: dict) -> dict:
        if not values["sender_alias"]:
            values["sender_alias"] = values["sender_login"]

        if not values["sender_login"]:
            logger.warning(
                "No login found for sending the email in action config. "
                "This will only work for email server that does not require authentication."
            )
        if not values["sender_password"]:
            logger.warning(
                "No password found for sending the email in action config."
                "This will only work for email server that does not require authentication."
            )

        return values

    @override
    def run(
        self,
        checkpoint_result: CheckpointResult,
        action_context: ActionContext | None = None,
    ) -> dict:
        success = checkpoint_result.success or False
        if not _should_notify(success=success, notify_on=self.notify_on):
            return {"email_result": ""}

        title, html = self.renderer.render(checkpoint_result=checkpoint_result)
        substituted_receiver_emails = (
            self._substitute_config_str_if_needed(self.receiver_emails) or ""
        )

        receiver_emails_list = list(
            map(lambda x: x.strip(), substituted_receiver_emails.split(","))
        )

        # this will actually send the email
        email_result = self._send_email(
            title=title,
            html=html,
            receiver_emails_list=receiver_emails_list,
        )

        checkpoint = checkpoint_result.checkpoint_config
        submit_event(
            event=NotificationActionRanEvent(
                type=self.type, notify_type=self.notify_on, checkpoint_id=checkpoint.id
            )
        )

        # sending payload back as dictionary
        return {"email_result": email_result}

    def _send_email(  # noqa: C901 # FIXME CoP
        self,
        title,
        html,
        receiver_emails_list,
    ):
        smtp_address = self._substitute_config_str_if_needed(self.smtp_address)
        smtp_port = self._substitute_config_str_if_needed(self.smtp_port)
        sender_login = self._substitute_config_str_if_needed(self.sender_login)
        sender_password = self._substitute_config_str_if_needed(self.sender_password)
        sender_alias = self._substitute_config_str_if_needed(self.sender_alias)

        msg = MIMEMultipart()
        msg["From"] = sender_alias
        msg["To"] = ", ".join(receiver_emails_list)
        msg["Subject"] = title
        msg.attach(MIMEText(html, "html"))
        try:
            if self.use_ssl:
                if self.use_tls:
                    logger.warning("Please choose between SSL or TLS, will default to SSL")
                context = ssl.create_default_context()
                mailserver = smtplib.SMTP_SSL(smtp_address, smtp_port, context=context)
            elif self.use_tls:
                mailserver = smtplib.SMTP(smtp_address, smtp_port)
                context = ssl.create_default_context()
                mailserver.starttls(context=context)
            else:
                logger.warning("Not using TLS or SSL to send an email is not secure")
                mailserver = smtplib.SMTP(smtp_address, smtp_port)
            if sender_login is not None and sender_password is not None:
                mailserver.login(sender_login, sender_password)
            elif not (sender_login is None and sender_password is None):
                logger.error(
                    "Please specify both sender_login and sender_password or specify both as None"
                )
            mailserver.sendmail(sender_alias, receiver_emails_list, msg.as_string())
            mailserver.quit()
        except smtplib.SMTPConnectError:
            logger.error(f"Failed to connect to the SMTP server at address: {smtp_address}")  # noqa: TRY400 # FIXME CoP
        except smtplib.SMTPAuthenticationError:
            logger.error(f"Failed to authenticate to the SMTP server at address: {smtp_address}")  # noqa: TRY400 # FIXME CoP
        except Exception as e:
            logger.error(str(e))  # noqa: TRY400 # FIXME CoP
        else:
            return "success"


@public_api
class UpdateDataDocsAction(DataDocsAction):
    """Notify the site builders of all data docs sites of a Data Context that a validation result should be added to the data docs.

    YAML configuration example:

    ```yaml
    - name: update_data_docs
    action:
      class_name: UpdateDataDocsAction
    ```

    You can also instruct ``UpdateDataDocsAction`` to build only certain sites by providing a ``site_names`` key with a
    list of sites to update:

    ```yaml
    - name: update_data_docs
    action:
      class_name: UpdateDataDocsAction
      site_names:
        - local_site
    ```

    Args:
        site_names: Optional. A list of the names of sites to update.
    """  # noqa: E501 # FIXME CoP

    type: Literal["update_data_docs"] = "update_data_docs"

    site_names: List[str] = []

    @override
    def run(
        self, checkpoint_result: CheckpointResult, action_context: ActionContext | None = None
    ) -> dict:
        action_results: dict[ValidationResultIdentifier, dict[str, str]] = {}
        for result_identifier, result in checkpoint_result.run_results.items():
            suite_name = result.suite_name

            expectation_suite_identifier: ExpectationSuiteIdentifier | GXCloudIdentifier
            if self._using_cloud_context:
                expectation_suite_identifier = GXCloudIdentifier(
                    resource_type=GXCloudRESTResource.EXPECTATION_SUITE, resource_name=suite_name
                )
            else:
                expectation_suite_identifier = ExpectationSuiteIdentifier(name=suite_name)

            action_result = self._run(
                validation_result_suite=result,
                validation_result_suite_identifier=result_identifier,
                expectation_suite_identifier=expectation_suite_identifier,
            )
            action_results[result_identifier] = action_result

        return action_results

    def _run(
        self,
        validation_result_suite: ExpectationSuiteValidationResult,
        validation_result_suite_identifier: Union[ValidationResultIdentifier, GXCloudIdentifier],
        action_context=None,
        expectation_suite_identifier=None,
        checkpoint_identifier=None,
    ):
        logger.debug("UpdateDataDocsAction.run")

        if validation_result_suite is None:
            logger.warning(
                f"No validation_result_suite was passed to {type(self).__name__} action. Skipping action."  # noqa: E501 # FIXME CoP
            )
            return

        if not isinstance(
            validation_result_suite_identifier,
            (ValidationResultIdentifier, GXCloudIdentifier),
        ):
            raise TypeError(  # noqa: TRY003 # FIXME CoP
                "validation_result_id must be of type ValidationResultIdentifier or"
                f" GeCloudIdentifier, not {type(validation_result_suite_identifier)}"
            )

        # TODO Update for RenderedDataDocs
        # build_data_docs will return the index page for the validation results, but we want to return the url for the validation result using the code below  # noqa: E501 # FIXME CoP
        self._build_data_docs(
            site_names=self.site_names,
            resource_identifiers=[
                validation_result_suite_identifier,
                expectation_suite_identifier,
            ],
        )
        data_docs_validation_results: dict = {}
        if self._using_cloud_context:
            return data_docs_validation_results

        # get the URL for the validation result
        docs_site_urls_list = self._get_docs_sites_urls(
            resource_identifier=validation_result_suite_identifier,
            site_names=self.site_names,
        )
        # process payload
        for sites in docs_site_urls_list:
            data_docs_validation_results[sites["site_name"]] = sites["site_url"]

        return data_docs_validation_results


class SNSNotificationAction(ValidationAction):
    """Action that pushes validations results to an SNS topic with a subject of passed or failed.

    YAML configuration example:

        ```yaml
        - name: send_sns_notification_on_validation_result
        action:
          class_name: SNSNotificationAction
          # put the actual SNS Arn in the uncommitted/config_variables.yml file
          # or pass in as environment variable
          sns_topic_arn:
          sns_subject:
        ```

    Args:
        sns_topic_arn: The SNS Arn to publish messages to.
        sns_subject: Optional. The SNS Message Subject - defaults to expectation_suite_identifier.name.
    """  # noqa: E501 # FIXME CoP

    type: Literal["sns"] = "sns"

    sns_topic_arn: str
    sns_message_subject: Optional[str]

    @override
    def run(
        self, checkpoint_result: CheckpointResult, action_context: ActionContext | None = None
    ) -> dict:
        msg = self._send_sns_notification(
            sns_subject=self.sns_message_subject or checkpoint_result.name,
            validation_results=json.dumps(
                [result.to_json_dict() for result in checkpoint_result.run_results.values()],
                indent=4,
            ),
        )
        return {"result": msg}

    def _send_sns_notification(self, sns_subject: str, validation_results: str, **kwargs) -> str:
        """
        Send JSON results to an SNS topic with a schema of:

        :param sns_subject: : The SNS Message Subject - defaults to suite_identifier.name
        :param validation_results:  The results of the validation ran
        :param kwargs:  Keyword arguments to pass to the boto3 Session
        :return:  Message ID that was published or error message

        """
        if not aws.boto3:
            logger.warning("boto3 is not installed")
            return "boto3 is not installed"

        message_dict = {
            "TopicArn": self.sns_topic_arn,
            "Subject": sns_subject,
            "Message": json.dumps(validation_results),
            "MessageAttributes": {
                "String": {"DataType": "String.Array", "StringValue": "ValidationResults"},
            },
            "MessageStructure": "json",
        }
        session = aws.boto3.Session(**kwargs)
        sns = session.client("sns")
        try:
            response = sns.publish(**message_dict)
        except sns.exceptions.InvalidParameterException:
            error_msg = f"Received invalid for message: {validation_results}"
            logger.error(error_msg)  # noqa: TRY400 # FIXME CoP
            return error_msg
        else:
            return (
                f"Successfully posted results to {response['MessageId']} with Subject {sns_subject}"
            )


class APINotificationAction(ValidationAction):
    type: Literal["api"] = "api"

    url: str

    @override
    def run(
        self, checkpoint_result: CheckpointResult, action_context: ActionContext | None = None
    ) -> dict:
        aggregate_payload = []
        for run_id, run_result in checkpoint_result.run_results.items():
            suite_name = run_result.suite_name
            serializable_results = convert_to_json_serializable(run_result.results)
            batch_identifier = run_id.batch_identifier

            payload = self.create_payload(
                data_asset_name=batch_identifier,
                suite_name=suite_name,
                validation_results_serializable=serializable_results,
            )
            aggregate_payload.append(payload)

        response = self.send_results(aggregate_payload)
        return {"result": f"Posted results to API, status code - {response.status_code}"}

    def send_results(self, payload) -> requests.Response:
        try:
            headers = {"Content-Type": "application/json"}
            return requests.post(self.url, headers=headers, data=payload)
        except Exception as e:
            print(f"Exception when sending data to API - {e}")
            raise e  # noqa: TRY201 # FIXME CoP

    @staticmethod
    def create_payload(data_asset_name, suite_name, validation_results_serializable) -> dict:
        return {
            "test_suite_name": suite_name,
            "data_asset_name": data_asset_name,
            "validation_results": validation_results_serializable,
        }



from __future__ import annotations

from uuid import UUID

import pytest

import great_expectations.expectations as gxe
from great_expectations.core.expectation_suite import ExpectationSuite
from great_expectations.data_context.data_context.cloud_data_context import CloudDataContext
from great_expectations.data_context.store import ExpectationsStore
from great_expectations.data_context.types.resource_identifiers import (
    ExpectationSuiteIdentifier,
    GXCloudIdentifier,
)
from great_expectations.util import gen_directory_tree_str


@pytest.mark.filesystem
def test_expectations_store():
    my_store = ExpectationsStore()

    with pytest.raises(TypeError):
        my_store.set("not_a_ValidationResultIdentifier")

    ns_1 = ExpectationSuiteIdentifier.from_tuple(tuple("a.b.c.warning"))
    my_store.set(
        ns_1,
        ExpectationSuite(name="a.b.c.warning"),
    )

    ns_1_dict: dict = my_store.get(ns_1)
    ns_1_suite = ExpectationSuite(**ns_1_dict)
    assert ns_1_suite == ExpectationSuite(name="a.b.c.warning")

    ns_2 = ExpectationSuiteIdentifier.from_tuple(tuple("a.b.c.failure"))
    my_store.set(
        ns_2,
        ExpectationSuite(name="a.b.c.failure"),
    )
    ns_2_dict: dict = my_store.get(ns_2)
    ns_2_suite = ExpectationSuite(**ns_2_dict)
    assert ns_2_suite == ExpectationSuite(name="a.b.c.failure")

    assert set(my_store.list_keys()) == {
        ns_1,
        ns_2,
    }


@pytest.mark.filesystem
def test_ExpectationsStore_with_DatabaseStoreBackend(sa):
    # Use sqlite so we don't require postgres for this test.
    connection_kwargs = {"drivername": "sqlite"}

    # First, demonstrate that we pick up default configuration
    my_store = ExpectationsStore(
        store_backend={
            "class_name": "DatabaseStoreBackend",
            "credentials": connection_kwargs,
        }
    )
    with pytest.raises(TypeError):
        my_store.get("not_a_ExpectationSuiteIdentifier")

    # first suite to add to db
    default_suite = ExpectationSuite(
        name="a.b.c.warning",
        meta={"test_meta_key": "test_meta_value"},
        expectations=[],
    )

    ns_1 = ExpectationSuiteIdentifier.from_tuple(tuple("a.b.c.warning"))
    # initial set and check if first suite exists
    my_store.set(ns_1, default_suite)
    ns_1_dict: dict = my_store.get(ns_1)
    ns_1_suite = ExpectationSuite(**ns_1_dict)
    assert ns_1_suite == ExpectationSuite(
        name="a.b.c.warning",
        meta={"test_meta_key": "test_meta_value"},
        expectations=[],
    )

    # update suite and check if new value exists
    updated_suite = ExpectationSuite(
        name="a.b.c.warning",
        meta={"test_meta_key": "test_new_meta_value"},
        expectations=[],
    )
    my_store.set(ns_1, updated_suite)
    ns_1_dict: dict = my_store.get(ns_1)
    ns_1_suite = ExpectationSuite(**ns_1_dict)
    assert ns_1_suite == ExpectationSuite(
        name="a.b.c.warning",
        meta={"test_meta_key": "test_new_meta_value"},
        expectations=[],
    )

    ns_2 = ExpectationSuiteIdentifier.from_tuple(tuple("a.b.c.failure"))
    my_store.set(
        ns_2,
        ExpectationSuite(name="a.b.c.failure"),
    )
    ns_2_dict: dict = my_store.get(ns_2)
    ns_2_suite = ExpectationSuite(**ns_2_dict)
    assert ns_2_suite == ExpectationSuite(
        name="a.b.c.failure",
    )

    assert set(my_store.list_keys()) == {
        ns_1,
        ns_2,
    }


@pytest.mark.unit
def test_expectations_store_report_store_backend_id_in_memory_store_backend():
    """
    What does this test and why?
    A Store should be able to report it's store_backend_id
    which is set when the StoreBackend is instantiated.
    """
    in_memory_expectations_store = ExpectationsStore()
    # Check that store_backend_id exists can be read
    assert in_memory_expectations_store.store_backend_id is not None
    # Check that store_backend_id is a valid UUID
    assert isinstance(in_memory_expectations_store.store_backend_id, UUID)


@pytest.mark.filesystem
def test_expectations_store_report_same_id_with_same_configuration_TupleFilesystemStoreBackend(
    tmp_path_factory,
):
    """
    What does this test and why?
    A store with the same config (must be persistent store) should report the same store_backend_id
    """
    full_test_dir = tmp_path_factory.mktemp(
        "test_expectations_store_report_same_id_with_same_configuration__dir"
    )
    test_dir = full_test_dir.parts[-1]
    project_path = str(full_test_dir)

    assert (
        gen_directory_tree_str(project_path)
        == f"""\
{test_dir}/
"""
    )

    # Check two stores with the same config
    persistent_expectations_store = ExpectationsStore(
        store_backend={
            "class_name": "TupleFilesystemStoreBackend",
            "base_directory": project_path,
        }
    )
    # Check successful initialization with a store_backend_id
    initialized_directory_tree_with_store_backend_id = f"""\
{test_dir}/
    .ge_store_backend_id
"""
    assert gen_directory_tree_str(project_path) == initialized_directory_tree_with_store_backend_id

    assert persistent_expectations_store.store_backend_id is not None

    # Check that a duplicate store reports the same store_backend_id
    persistent_expectations_store_duplicate = ExpectationsStore(
        store_backend={
            "class_name": "TupleFilesystemStoreBackend",
            "base_directory": project_path,
        }
    )
    assert persistent_expectations_store_duplicate.store_backend_id is not None
    assert (
        persistent_expectations_store.store_backend_id
        == persistent_expectations_store_duplicate.store_backend_id
    )
    # Check no change to filesystem
    assert gen_directory_tree_str(project_path) == initialized_directory_tree_with_store_backend_id


def _create_suite_config(name: str, id: str, expectations: list[dict] | None = None) -> dict:
    return {
        "id": id,
        "name": name,
        "expectations": expectations or [],
        "meta": {},
        "notes": None,
    }


_SUITE_CONFIG = _create_suite_config("my_suite", "03d61d4e-003f-48e7-a3b2-f9f842384da3")
_SUITE_CONFIG_WITH_EXPECTATIONS = _create_suite_config(
    "my_suite_with_expectations",
    "03d61d4e-003f-48e7-a3b2-f9f842384da3",
    [
        {
            "type": "expect_column_to_exist",
            "description": None,
            "id": "c8a239a6-fb80-4f51-a90e-40c38dffdf91",
            "kwargs": {"column": "infinities"},
            "meta": {},
            "expectation_context": None,
            "rendered_content": [],
        }
    ],
)

# this should have non-null values
_SUITE_CONFIG_WITH_FULLY_POPULATED_EXPECTATIONS = _create_suite_config(
    "my_suite_with_expectations",
    "03d61d4e-003f-48e7-a3b2-f9f842384da3",
    [
        {
            "type": "expect_column_to_exist",
            "description": "some description",
            "id": "c8a239a6-fb80-4f51-a90e-40c38dffdf91",
            "kwargs": {"column": "infinities"},
            "meta": {},
            "expectation_context": None,
            "rendered_content": [],
        }
    ],
)


@pytest.mark.unit
@pytest.mark.parametrize(
    "response_json, expected, error_type",
    [
        pytest.param(
            {"data": _SUITE_CONFIG},
            _SUITE_CONFIG,
            None,
            id="single_config",
        ),
        pytest.param({"data": []}, None, ValueError, id="empty_payload"),
        pytest.param(
            {"data": [_SUITE_CONFIG]},
            _SUITE_CONFIG,
            None,
            id="single_config_in_list",
        ),
        pytest.param(
            {"data": _SUITE_CONFIG_WITH_EXPECTATIONS},
            _SUITE_CONFIG_WITH_EXPECTATIONS,
            None,
            id="null_result_format",
        ),
        pytest.param(
            {"data": _SUITE_CONFIG_WITH_FULLY_POPULATED_EXPECTATIONS},
            _SUITE_CONFIG_WITH_FULLY_POPULATED_EXPECTATIONS,
            None,
            id="null_result_format",
        ),
    ],
)
def test_gx_cloud_response_json_to_object_dict(
    response_json: dict, expected: dict | None, error_type: type[Exception] | None
) -> None:
    if error_type:
        with pytest.raises(error_type):
            _ = ExpectationsStore.gx_cloud_response_json_to_object_dict(response_json)
    else:
        actual = ExpectationsStore.gx_cloud_response_json_to_object_dict(response_json)
        assert actual == expected


@pytest.mark.unit
def test_gx_cloud_response_json_to_object_collection():
    response_json = {
        "data": [
            _SUITE_CONFIG,
            _SUITE_CONFIG_WITH_EXPECTATIONS,
        ]
    }

    result = ExpectationsStore.gx_cloud_response_json_to_object_collection(response_json)

    expected = [_SUITE_CONFIG, _SUITE_CONFIG_WITH_EXPECTATIONS]
    assert result == expected


@pytest.mark.unit
def test_get_key_in_non_cloud_mode(empty_data_context):
    name = "test-name"
    suite = ExpectationSuite(name=name)
    key = empty_data_context.expectations_store.get_key(name=suite.name, id=suite.id)
    assert isinstance(key, ExpectationSuiteIdentifier)
    assert key.name == name


@pytest.mark.unit
def test_get_key_in_cloud_mode(empty_data_context_in_cloud_mode):
    cloud_data_context = empty_data_context_in_cloud_mode
    name = "test-name"
    suite = ExpectationSuite(name=name)
    key = cloud_data_context.expectations_store.get_key(name=suite.name, id=suite.id)
    assert isinstance(key, GXCloudIdentifier)
    assert key.resource_name == name


@pytest.mark.cloud
def test_add_expectation_success_cloud_backend(empty_cloud_data_context):
    context = empty_cloud_data_context
    _test_add_expectation_success(context)


@pytest.mark.filesystem
def test_add_expectation_success_filesystem_backend(empty_data_context):
    context = empty_data_context
    _test_add_expectation_success(context)


def _test_add_expectation_success(context):
    # Arrange
    store = context.expectations_store
    suite_name = "test-suite"
    suite = ExpectationSuite(suite_name)
    context.suites.add(suite)
    expectation = gxe.ExpectColumnValuesToBeInSet(
        column="a",
        value_set=[1, 2, 3],
        result_format="BASIC",
    )
    # Act
    store.add_expectation(suite=suite, expectation=expectation)
    # Assert
    updated_suite_dict = store.get(key=store.get_key(name=suite.name, id=suite.id))
    updated_suite = ExpectationSuite(**updated_suite_dict)
    added_expectation = updated_suite.expectations[0]
    assert UUID(added_expectation.id)
    assert expectation.column == added_expectation.column
    assert expectation.value_set == added_expectation.value_set
    assert expectation.result_format == added_expectation.result_format


@pytest.mark.filesystem
def test_add_expectation_disregards_provided_id_filesystem_backend(empty_data_context):
    context = empty_data_context
    _test_add_expectation_disregards_provided_id(context)


@pytest.mark.cloud
def test_add_expectation_disregards_provided_id_cloud_backend(empty_cloud_data_context):
    context = empty_cloud_data_context
    _test_add_expectation_disregards_provided_id(context)


def _test_add_expectation_disregards_provided_id(context):
    # Arrange
    store = context.expectations_store
    suite_name = "test-suite"
    suite = ExpectationSuite(suite_name)
    context.suites.add(suite)
    provided_id = "e86bb8a8-b75f-4efb-a3bb-210b6440661e"
    expectation = gxe.ExpectColumnValuesToBeInSet(
        id=provided_id,
        column="a",
        value_set=[1, 2, 3],
        result_format="BASIC",
    )
    # Act
    store.add_expectation(suite=suite, expectation=expectation)
    # Assert
    updated_suite_dict = store.get(key=store.get_key(name=suite.name, id=suite.id))
    updated_suite = ExpectationSuite(**updated_suite_dict)
    added_expectation = updated_suite.expectations[0]
    assert UUID(added_expectation.id)
    assert added_expectation.id != provided_id


@pytest.mark.filesystem
def test_add_adds_ids_to_suite_and_expectations(empty_data_context):
    context = empty_data_context

    expectation_a = gxe.ExpectColumnValuesToBeInSet(column="a", value_set=[1, 2, 3])
    expectation_b = gxe.ExpectColumnMaxToBeBetween(column="b", min_value=0, max_value=10)
    suite = ExpectationSuite("test-suite", expectations=[expectation_a, expectation_b])

    assert all(obj.id is None for obj in (expectation_a, expectation_b, suite))

    suite = context.suites.add(suite)

    assert all(obj.id is not None for obj in (expectation_a, expectation_b, suite))


@pytest.mark.cloud
def test_update_expectation_success_cloud_backend(empty_cloud_data_context):
    context = empty_cloud_data_context
    _test_update_expectation_success(context)


@pytest.mark.filesystem
def test_update_expectation_success_file_backend(empty_data_context):
    context = empty_data_context
    _test_update_expectation_success(context)


def _test_update_expectation_success(context):
    # Arrange
    store = context.expectations_store
    suite_name = "test-suite"
    expectation = gxe.ExpectColumnValuesToBeInSet(
        column="a",
        value_set=[1, 2, 3],
        result_format="BASIC",
    )
    suite = ExpectationSuite(suite_name, expectations=[expectation.configuration])
    context.suites.add(suite)
    # Act
    expectation = suite.expectations[0]
    assert expectation.column == "a"
    updated_column_name = "foo"
    expectation.column = updated_column_name
    store.update_expectation(suite=suite, expectation=expectation)
    # Assert
    updated_suite_dict = store.get(key=store.get_key(name=suite.name, id=suite.id))
    updated_suite = ExpectationSuite(**updated_suite_dict)
    updated_expectation = updated_suite.expectations[0]
    assert updated_expectation.id == expectation.id
    assert updated_expectation.column == updated_column_name


@pytest.mark.filesystem
def test_update_expectation_raises_error_for_missing_expectation_filesystem(
    empty_data_context,
):
    # Arrange
    context = empty_data_context
    _test_update_expectation_raises_error_for_missing_expectation(context)


@pytest.mark.cloud
def test_update_expectation_raises_error_for_missing_expectation_cloud(
    empty_cloud_data_context,
):
    # Arrange
    context = empty_cloud_data_context
    _test_update_expectation_raises_error_for_missing_expectation(context)


def _test_update_expectation_raises_error_for_missing_expectation(context):
    store = context.expectations_store
    suite_name = "test-suite"
    expectation = gxe.ExpectColumnValuesToBeInSet(
        id="2b284004-0e0e-455d-a7f4-11e162fd06c9",
        column="a",
        value_set=[1, 2, 3],
        result_format="BASIC",
    )

    suite = ExpectationSuite(suite_name, expectations=[])
    context.suites.add(suite)
    # Act
    with pytest.raises(KeyError, match="Cannot update Expectation because it was not found."):
        store.update_expectation(suite=suite, expectation=expectation)
    # Assert
    updated_suite_dict = store.get(key=store.get_key(name=suite.name, id=suite.id))
    updated_suite = ExpectationSuite(**updated_suite_dict)
    assert suite == updated_suite


@pytest.mark.cloud
def test_delete_expectation_success_cloud_backend(empty_cloud_data_context):
    # Arrange
    context = empty_cloud_data_context
    _test_delete_expectation_success(context)


@pytest.mark.filesystem
def test_delete_expectation_success_filesystem_backend(empty_data_context):
    # Arrange
    context = empty_data_context
    _test_delete_expectation_success(context)


def _test_delete_expectation_success(context):
    store = context.expectations_store
    suite_name = "test-suite"
    expectation = gxe.ExpectColumnValuesToBeInSet(
        column="a",
        value_set=[1, 2, 3],
        result_format="BASIC",
    )
    suite = ExpectationSuite(suite_name, expectations=[expectation.configuration])
    context.suites.add(suite)
    # Act
    expectation = suite.expectations[0]
    store.delete_expectation(suite=suite, expectation=expectation)
    # Assert
    updated_suite_dict = store.get(key=store.get_key(name=suite.name, id=suite.id))
    updated_suite = ExpectationSuite(**updated_suite_dict)
    assert len(updated_suite.expectations) == 0


@pytest.mark.filesystem
def test_delete_expectation_raises_error_for_missing_expectation_filesystem(
    empty_data_context,
):
    context = empty_data_context
    _test_delete_expectation_raises_error_for_missing_expectation(context)


@pytest.mark.cloud
def test_delete_expectation_raises_error_for_missing_expectation_cloud(
    empty_cloud_data_context,
):
    context = empty_cloud_data_context
    _test_delete_expectation_raises_error_for_missing_expectation(context)


def _test_delete_expectation_raises_error_for_missing_expectation(context):
    # Arrange
    store = context.expectations_store
    suite_name = "test-suite"
    existing_expectation = gxe.ExpectColumnValuesToBeInSet(
        column="a",
        value_set=[1, 2, 3],
        result_format="BASIC",
    )
    suite = ExpectationSuite(suite_name, expectations=[existing_expectation.configuration])
    context.suites.add(suite)
    # Act
    nonexistent_expectation = gxe.ExpectColumnValuesToBeInSet(
        # this ID will be different from the ID created by the Suite
        id="1296c5c8-6f7b-4cee-a09c-9037c7e40df7",
        column="a",
        value_set=[1, 2, 3],
        result_format="BASIC",
    )
    with pytest.raises(KeyError, match="Cannot delete Expectation because it was not found."):
        store.delete_expectation(suite=suite, expectation=nonexistent_expectation)
    # Assert
    updated_suite_dict = store.get(key=store.get_key(name=suite.name, id=suite.id))
    updated_suite = ExpectationSuite(**updated_suite_dict)
    if isinstance(context, CloudDataContext):
        updated_suite.render()
    assert suite == updated_suite
    assert len(updated_suite.expectations) == 1


@pytest.mark.cloud
def test_update_cloud_suite(empty_cloud_data_context):
    # Arrange
    context = empty_cloud_data_context
    store = context.expectations_store
    existing_expectation = gxe.ExpectColumnValuesToBeInSet(
        column="a",
        value_set=[1, 2, 3],
        result_format="BASIC",
    )
    suite_name = "test-suite"
    updated_suite_name = "test-suite-22"
    suite = ExpectationSuite(suite_name, expectations=[existing_expectation.configuration])
    context.suites.add(suite)

    # act
    suite.name = updated_suite_name
    store.update(key=store.get_key(name=suite_name, id=suite.id), value=suite)

    # assert
    updated_suite_dict = store.get(key=store.get_key(name=updated_suite_name, id=suite.id))
    assert updated_suite_dict["name"] == updated_suite_name



from __future__ import annotations

import logging
from abc import ABCMeta
from typing import TYPE_CHECKING, Any, Callable, List, Literal, Protocol

from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.id_dict import BatchSpec
from great_expectations.exceptions import InvalidBatchIdError, InvalidBatchSpecError
from great_expectations.types.base import SerializableDotDict

if TYPE_CHECKING:
    import pandas as pd
    from typing_extensions import TypeAlias

    from great_expectations.alias_types import JSONValues, PathStr

logger = logging.getLogger(__name__)


# TODO: <Alex>This module needs to be cleaned up.
#  We have Batch used for the legacy design, and we also need Batch for the new design.
#  However, right now, the Batch from the legacy design is imported into execution engines of the new design.  # noqa: E501 # FIXME CoP
#  As a result, we have multiple, inconsistent versions of BatchMarkers, extending legacy/new classes.</Alex>  # noqa: E501 # FIXME CoP
# TODO: <Alex>See also "great_expectations/core/batch.py".</Alex>
# TODO: <Alex>The following class is part of the new design.</Alex>
class BatchMarkers(BatchSpec):
    """A BatchMarkers is a special type of BatchSpec (so that it has a batch_fingerprint) but it generally does
    NOT require specific keys and instead captures information about the OUTPUT of a datasource's fetch
    process, such as the timestamp at which a query was executed."""  # noqa: E501 # FIXME CoP

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        if "ge_load_time" not in self:
            raise InvalidBatchIdError("BatchMarkers requires a ge_load_time")  # noqa: TRY003 # FIXME CoP

    @property
    def ge_load_time(self):
        return self.get("ge_load_time")


class PandasBatchSpecProtocol(Protocol):
    @property
    def reader_method(self) -> str: ...

    @property
    def reader_options(self) -> dict: ...

    def to_json_dict(self) -> dict[str, JSONValues]: ...


class PandasBatchSpec(SerializableDotDict, BatchSpec, PandasBatchSpecProtocol):
    @property
    @override
    def reader_method(self) -> str:
        return self["reader_method"]

    @property
    @override
    def reader_options(self) -> dict:
        return self.get("reader_options", {})

    @override
    def to_json_dict(self) -> dict[str, JSONValues]:
        from great_expectations.datasource.fluent.pandas_datasource import (
            _EXCLUDE_TYPES_FROM_JSON,
        )

        json_dict: dict[str, JSONValues] = dict()
        json_dict["reader_method"] = self.reader_method
        json_dict["reader_options"] = {
            reader_option_name: reader_option
            for reader_option_name, reader_option in self.reader_options.items()
            if not isinstance(reader_option, tuple(_EXCLUDE_TYPES_FROM_JSON))
        }
        return json_dict


class PathBatchSpec(BatchSpec, metaclass=ABCMeta):
    def __init__(
        self,
        *args,
        path: PathStr = None,  # type: ignore[assignment] # error raised if not provided
        reader_options: dict[str, Any] | None = None,
        **kwargs,
    ) -> None:
        if path:
            kwargs["path"] = str(path)
        if reader_options:
            kwargs["reader_options"] = reader_options
        super().__init__(*args, **kwargs)
        if "path" not in self:
            raise InvalidBatchSpecError("PathBatchSpec requires a path element")  # noqa: TRY003 # FIXME CoP

    @property
    def path(self) -> str:
        return self.get("path")  # type: ignore[return-value] # FIXME CoP

    @property
    def reader_method(self) -> str:
        return self.get("reader_method")  # type: ignore[return-value] # FIXME CoP

    @property
    def reader_options(self) -> dict:
        return self.get("reader_options") or {}


FabricReaderMethods: TypeAlias = Literal["read_table", "evaluate_measure", "evaluate_dax"]


class FabricBatchSpec(PandasBatchSpecProtocol):
    # TODO: use slots

    def __init__(
        self,
        reader_method: FabricReaderMethods,
        reader_options: dict[str, Any],
    ) -> None:
        self._reader_method = reader_method
        self._reader_options = reader_options

    @property
    @override
    def reader_method(self) -> str:
        return self._reader_method

    @property
    @override
    def reader_options(self) -> dict[str, Any]:
        return self._reader_options

    @override
    def to_json_dict(self) -> dict[str, JSONValues]:
        return {
            "reader_method": self.reader_method,
            "reader_options": self.reader_options,
        }

    def get_reader_function(self) -> Callable[..., pd.DataFrame]:
        # lazy import of fabric module which contains the reader functions
        from sempy import fabric

        try:
            return getattr(fabric, self.reader_method)
        except AttributeError:
            raise AttributeError(  # noqa: TRY003 # FIXME CoP
                f"FabricBatchSpec reader_method {self.reader_method} not found in sempy.fabric module"  # noqa: E501 # FIXME CoP
            )


class S3BatchSpec(PathBatchSpec):
    pass


class AzureBatchSpec(PathBatchSpec):
    pass


class GCSBatchSpec(PathBatchSpec):
    pass


class SqlAlchemyDatasourceBatchSpec(BatchSpec, metaclass=ABCMeta):
    """This is an abstract class and should not be instantiated. It's relevant for testing whether
    a subclass is allowed
    """

    @property
    def limit(self):
        return self.get("limit")

    @property
    def schema(self):
        return self.get("schema")


class RuntimeDataBatchSpec(BatchSpec):
    _id_ignore_keys = set("batch_data")

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        if self.batch_data is None:
            raise InvalidBatchSpecError("RuntimeDataBatchSpec batch_data cannot be None")  # noqa: TRY003 # FIXME CoP

    @property
    def batch_data(self):
        return self.get("batch_data")

    @batch_data.setter
    def batch_data(self, batch_data) -> None:
        self["batch_data"] = batch_data


class RuntimeQueryBatchSpec(BatchSpec):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        if self.query is None:
            raise InvalidBatchSpecError("RuntimeQueryBatchSpec query cannot be None")  # noqa: TRY003 # FIXME CoP

    @property
    def query(self):
        return self.get("query")

    @query.setter
    def query(self, query) -> None:
        self["query"] = query


class GlueDataCatalogBatchSpec(BatchSpec):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        if "database_name" not in self:
            raise InvalidBatchSpecError("GlueDataCatalogBatchSpec requires a database_name")  # noqa: TRY003 # FIXME CoP
        if "table_name" not in self:
            raise InvalidBatchSpecError("GlueDataCatalogBatchSpec requires a table_name")  # noqa: TRY003 # FIXME CoP

    @property
    def reader_method(self) -> str:
        return "table"

    @property
    def database_name(self) -> str:
        return self["database_name"]

    @property
    def table_name(self) -> str:
        return self["table_name"]

    @property
    def path(self) -> str:
        return f"{self.database_name}.{self.table_name}"

    @property
    def reader_options(self) -> dict:
        return self.get("reader_options", {})

    @property
    def partitions(self) -> List[str]:
        return self.get("partitions", [])



import pandas as pd

from great_expectations.metrics.batch.row_count import BatchRowCount, BatchRowCountResult
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.metrics.conftest import PANDAS_DATA_SOURCES, SPARK_DATA_SOURCES, SQL_DATA_SOURCES

DATA_FRAME = pd.DataFrame(
    {
        "id": [1, 2, 3, 4],
        "name": [1, 2, 3, 4],
    },
)


class TestBatchRowCount:
    ROW_COUNT = 4

    @parameterize_batch_for_data_sources(
        data_source_configs=PANDAS_DATA_SOURCES,
        data=DATA_FRAME,
    )
    def test_success_pandas(self, batch_for_datasource) -> None:
        batch = batch_for_datasource
        metric = BatchRowCount()
        metric_result = batch.compute_metrics(metric)
        assert isinstance(metric_result, BatchRowCountResult)
        assert metric_result.value == self.ROW_COUNT

    @parameterize_batch_for_data_sources(
        data_source_configs=SPARK_DATA_SOURCES,
        data=DATA_FRAME,
    )
    def test_success_spark(self, batch_for_datasource) -> None:
        batch = batch_for_datasource
        metric = BatchRowCount()
        metric_result = batch.compute_metrics(metric)
        assert isinstance(metric_result, BatchRowCountResult)
        assert metric_result.value == self.ROW_COUNT

    @parameterize_batch_for_data_sources(
        data_source_configs=SQL_DATA_SOURCES,
        data=DATA_FRAME,
    )
    def test_success_sql(self, batch_for_datasource) -> None:
        batch = batch_for_datasource
        metric = BatchRowCount()
        metric_result = batch.compute_metrics(metric)
        assert isinstance(metric_result, BatchRowCountResult)
        assert metric_result.value == self.ROW_COUNT



"""Create queries for use in sql data partitioning.

This module contains utilities for generating queries used by execution engines
and data connectors to partition data into batches based on the data itself. It
is typically used from within either an execution engine or a data connector,
not by itself.

    Typical usage example:
        __init__():
            self._sqlalchemy_data_partitioner = SqlAlchemyDataPartitioner()

        elsewhere():
            partitioner = self._sqlalchemy_data_partitioner._get_partitioner_method()
            partition_query_or_clause = partitioner()
"""

from __future__ import annotations

from typing import TYPE_CHECKING, List, Union

import great_expectations.exceptions as gx_exceptions
from great_expectations.compatibility.sqlalchemy import (
    sqlalchemy as sa,
)
from great_expectations.execution_engine.partition_and_sample.data_partitioner import (
    DataPartitioner,
    DatePart,
    PartitionerMethod,
)
from great_expectations.execution_engine.sqlalchemy_dialect import GXSqlDialect

if TYPE_CHECKING:
    from great_expectations.compatibility import sqlalchemy
    from great_expectations.execution_engine.sqlalchemy_execution_engine import (
        SqlAlchemyExecutionEngine,
    )


class SqlAlchemyDataPartitioner(DataPartitioner):
    """Methods for partitioning data accessible via SqlAlchemyExecutionEngine.

    Note, for convenience, you can also access DatePart via the instance variable
    date_part e.g. SqlAlchemyDataPartitioner.date_part.MONTH
    """

    def __init__(self, dialect: str):
        self._dialect = dialect

    DATETIME_PARTITIONER_METHOD_TO_GET_UNIQUE_BATCH_IDENTIFIERS_METHOD_MAPPING: dict = {
        PartitionerMethod.PARTITION_ON_YEAR: "get_data_for_batch_identifiers_year",
        PartitionerMethod.PARTITION_ON_YEAR_AND_MONTH: "get_data_for_batch_identifiers_year_and_month",  # noqa: E501 # FIXME CoP
        PartitionerMethod.PARTITION_ON_YEAR_AND_MONTH_AND_DAY: "get_data_for_batch_identifiers_year_and_month_and_day",  # noqa: E501 # FIXME CoP
        PartitionerMethod.PARTITION_ON_DATE_PARTS: "get_data_for_batch_identifiers_for_partition_on_date_parts",  # noqa: E501 # FIXME CoP
    }

    PARTITIONER_METHOD_TO_GET_UNIQUE_BATCH_IDENTIFIERS_METHOD_MAPPING: dict = {
        PartitionerMethod.PARTITION_ON_WHOLE_TABLE: "get_partition_query_for_data_for_batch_identifiers_for_partition_on_whole_table",  # noqa: E501 # FIXME CoP
        PartitionerMethod.PARTITION_ON_COLUMN_VALUE: "get_partition_query_for_data_for_batch_identifiers_for_partition_on_column_value",  # noqa: E501 # FIXME CoP
        PartitionerMethod.PARTITION_ON_CONVERTED_DATETIME: "get_partition_query_for_data_for_batch_identifiers_for_partition_on_converted_datetime",  # noqa: E501 # FIXME CoP
        PartitionerMethod.PARTITION_ON_DIVIDED_INTEGER: "get_partition_query_for_data_for_batch_identifiers_for_partition_on_divided_integer",  # noqa: E501 # FIXME CoP
        PartitionerMethod.PARTITION_ON_MOD_INTEGER: "get_partition_query_for_data_for_batch_identifiers_for_partition_on_mod_integer",  # noqa: E501 # FIXME CoP
        PartitionerMethod.PARTITION_ON_MULTI_COLUMN_VALUES: "get_partition_query_for_data_for_batch_identifiers_for_partition_on_multi_column_values",  # noqa: E501 # FIXME CoP
        PartitionerMethod.PARTITION_ON_HASHED_COLUMN: "get_partition_query_for_data_for_batch_identifiers_for_partition_on_hashed_column",  # noqa: E501 # FIXME CoP
    }

    def partition_on_year(
        self,
        column_name: str,
        batch_identifiers: dict,
    ) -> Union[sqlalchemy.BinaryExpression, sqlalchemy.BooleanClauseList]:
        """Partition on year values in column_name.

        Args:
            column_name: column in table to use in determining partition.
            batch_identifiers: should contain a dateutil parseable datetime whose
                relevant date parts will be used for partitioning or key values
                of {date_part: date_part_value}.

        Returns:
            List of boolean clauses based on whether the date_part value in the
                batch identifier matches the date_part value in the column_name column.
        """
        return self.partition_on_date_parts(
            column_name=column_name,
            batch_identifiers=batch_identifiers,
            date_parts=[DatePart.YEAR],
        )

    def partition_on_year_and_month(
        self,
        column_name: str,
        batch_identifiers: dict,
    ) -> Union[sqlalchemy.BinaryExpression, sqlalchemy.BooleanClauseList]:
        """Partition on year and month values in column_name.

        Args:
            column_name: column in table to use in determining partition.
            batch_identifiers: should contain a dateutil parseable datetime whose
                relevant date parts will be used for partitioning or key values
                of {date_part: date_part_value}.

        Returns:
            List of boolean clauses based on whether the date_part value in the
                batch identifier matches the date_part value in the column_name column.
        """
        return self.partition_on_date_parts(
            column_name=column_name,
            batch_identifiers=batch_identifiers,
            date_parts=[DatePart.YEAR, DatePart.MONTH],
        )

    def partition_on_year_and_month_and_day(
        self,
        column_name: str,
        batch_identifiers: dict,
    ) -> Union[sqlalchemy.BinaryExpression, sqlalchemy.BooleanClauseList]:
        """Partition on year and month and day values in column_name.

        Args:
            column_name: column in table to use in determining partition.
            batch_identifiers: should contain a dateutil parseable datetime whose
                relevant date parts will be used for partitioning or key values
                of {date_part: date_part_value}.

        Returns:
            List of boolean clauses based on whether the date_part value in the
                batch identifier matches the date_part value in the column_name column.
        """
        return self.partition_on_date_parts(
            column_name=column_name,
            batch_identifiers=batch_identifiers,
            date_parts=[DatePart.YEAR, DatePart.MONTH, DatePart.DAY],
        )

    def partition_on_date_parts(
        self,
        column_name: str,
        batch_identifiers: dict,
        date_parts: Union[List[DatePart], List[str]],
    ) -> Union[sqlalchemy.BinaryExpression, sqlalchemy.BooleanClauseList]:
        """Partition on date_part values in column_name.

        Values are NOT truncated, for example this will return data for a
        given month (if only month is chosen for date_parts) for ALL years.
        This may be useful for viewing seasonality, but you can also specify
        multiple date_parts to achieve date_trunc like behavior e.g.
        year, month and day.

        Args:
            column_name: column in table to use in determining partition.
            batch_identifiers: should contain a dateutil parseable datetime whose date parts
                will be used for partitioning or key values of {date_part: date_part_value}
            date_parts: part of the date to be used for partitioning e.g.
                DatePart.DAY or the case-insensitive string representation "day"

        Returns:
            List of boolean clauses based on whether the date_part value in the
                batch identifier matches the date_part value in the column_name column.
        """
        self._validate_date_parts(date_parts)

        date_parts = self._convert_date_parts(date_parts)

        column_batch_identifiers: dict = batch_identifiers[column_name]

        date_parts_dict: dict = self._convert_datetime_batch_identifiers_to_date_parts_dict(
            column_batch_identifiers, date_parts
        )

        query: Union[sqlalchemy.BinaryExpression, sqlalchemy.BooleanClauseList] = sa.and_(  # type: ignore[assignment] # FIXME CoP
            *[
                sa.extract(date_part.value, sa.column(column_name))
                == date_parts_dict[date_part.value]
                for date_part in date_parts
            ]
        )

        return query

    @staticmethod
    def partition_on_whole_table(batch_identifiers: dict) -> bool:
        """'Partition' by returning the whole table"""

        return True

    @staticmethod
    def partition_on_column_value(column_name: str, batch_identifiers: dict) -> bool:
        """Partition using the values in the named column"""

        return sa.column(column_name) == batch_identifiers[column_name]

    def partition_on_converted_datetime(
        self,
        column_name: str,
        batch_identifiers: dict,
        date_format_string: str = "%Y-%m-%d",
    ) -> bool:
        """Convert the values in the named column to the given date_format, and partition on that"""
        if self._dialect == GXSqlDialect.SQLITE:
            return (
                sa.func.strftime(
                    date_format_string,
                    sa.column(column_name),
                )
                == batch_identifiers[column_name]
            )

        raise NotImplementedError(
            f'Partitioner method "partition_on_converted_datetime" is not supported for "{self._dialect}" SQL dialect.'  # noqa: E501 # FIXME CoP
        )

    def partition_on_divided_integer(
        self,
        column_name: str,
        divisor: int,
        batch_identifiers: dict,
    ) -> bool:
        """Divide the values in the named column by `divisor`, and partition on that"""
        if self._dialect == GXSqlDialect.SQLITE:
            return (
                sa.cast((sa.cast(sa.column(column_name), sa.Integer) / divisor), sa.Integer)
                == batch_identifiers[column_name]
            )

        if self._dialect == GXSqlDialect.MYSQL:
            return (
                sa.cast(
                    sa.func.truncate((sa.cast(sa.column(column_name), sa.Integer) / divisor), 0),
                    sa.Integer,
                )
                == batch_identifiers[column_name]
            )

        if self._dialect == GXSqlDialect.MSSQL:
            return (
                sa.cast(
                    sa.func.round((sa.cast(sa.column(column_name), sa.Integer) / divisor), 0, 1),
                    sa.Integer,
                )
                == batch_identifiers[column_name]
            )

        if self._dialect == GXSqlDialect.AWSATHENA:
            return (
                sa.cast(
                    sa.func.truncate(sa.cast(sa.column(column_name), sa.Integer) / divisor),
                    sa.Integer,
                )
                == batch_identifiers[column_name]
            )

        return (
            sa.cast(
                sa.func.trunc((sa.cast(sa.column(column_name), sa.Integer) / divisor), 0),
                sa.Integer,
            )
            == batch_identifiers[column_name]
        )

    def partition_on_mod_integer(
        self,
        column_name: str,
        mod: int,
        batch_identifiers: dict,
    ) -> bool:
        """Divide the values in the named column by `mod`, and partition on that"""
        if self._dialect in [
            GXSqlDialect.SQLITE,
            GXSqlDialect.MSSQL,
        ]:
            return (
                sa.cast(sa.column(column_name), sa.Integer) % mod == batch_identifiers[column_name]
            )

        return (
            sa.func.mod(sa.cast(sa.column(column_name), sa.Integer), mod)
            == batch_identifiers[column_name]
        )

    @staticmethod
    def partition_on_multi_column_values(
        column_names: List[str],
        batch_identifiers: dict,
    ) -> bool:
        """Partition on the joint values in the named columns"""

        return sa.and_(  # type: ignore[return-value] # FIXME CoP
            *(
                sa.column(column_name) == column_value
                for column_name, column_value in batch_identifiers.items()
            )
        )

    def partition_on_hashed_column(
        self,
        column_name: str,
        hash_digits: int,
        batch_identifiers: dict,
    ) -> bool:
        """Partition on the hashed value of the named column"""
        if self._dialect == GXSqlDialect.SQLITE:
            return (
                sa.func.md5(sa.cast(sa.column(column_name), sa.VARCHAR), hash_digits)
                == batch_identifiers[column_name]
            )

        raise NotImplementedError(
            f'Partitioner method "partition_on_hashed_column" is not supported for "{self._dialect}" SQL dialect.'  # noqa: E501 # FIXME CoP
        )

    def get_data_for_batch_identifiers(
        self,
        execution_engine: SqlAlchemyExecutionEngine,
        selectable: sqlalchemy.Selectable,
        partitioner_method_name: str,
        partitioner_kwargs: dict,
    ) -> List[dict]:
        """Build data used to construct batch identifiers for the input table using the provided partitioner config.

        Sql partitioner configurations yield the unique values that comprise a batch by introspecting your data.

        Args:
            execution_engine: Used to introspect the data.
            selectable: Selectable to partition.
            partitioner_method_name: Desired partitioner method to use.
            partitioner_kwargs: Dict of directives used by the partitioner method as keyword arguments of key=value.

        Returns:
            List of dicts of the form [{column_name: {"key": value}}]
        """  # noqa: E501 # FIXME CoP
        processed_partitioner_method_name: str = self._get_partitioner_method_name(
            partitioner_method_name
        )
        batch_identifiers_list: List[dict]
        if self._is_datetime_partitioner(processed_partitioner_method_name):
            partitioner_fn_name: str = (
                self.DATETIME_PARTITIONER_METHOD_TO_GET_UNIQUE_BATCH_IDENTIFIERS_METHOD_MAPPING[
                    processed_partitioner_method_name
                ]
            )
            batch_identifiers_list = getattr(self, partitioner_fn_name)(
                execution_engine, selectable, **partitioner_kwargs
            )
        else:
            batch_identifiers_list = (
                self.get_data_for_batch_identifiers_for_non_date_part_partitioners(
                    execution_engine,
                    selectable,
                    processed_partitioner_method_name,
                    partitioner_kwargs,
                )
            )

        return batch_identifiers_list

    def _is_datetime_partitioner(self, partitioner_method_name: str) -> bool:
        """Whether the partitioner method is a datetime partitioner.

        Args:
            partitioner_method_name: Name of the partitioner method

        Returns:
            Boolean
        """
        return partitioner_method_name in list(
            self.DATETIME_PARTITIONER_METHOD_TO_GET_UNIQUE_BATCH_IDENTIFIERS_METHOD_MAPPING.keys()
        )

    def get_data_for_batch_identifiers_year(
        self,
        execution_engine: SqlAlchemyExecutionEngine,
        selectable: sqlalchemy.Selectable,
        column_name: str,
    ) -> List[dict]:
        """Build batch_identifiers from a column partition on year.

        This method builds a query to select the unique date_parts from the
        column_name. This data can be used to build BatchIdentifiers.

        Args:
            execution_engine: SqlAlchemyExecutionEngine to be used for executing the query.
            selectable: selectable to partition.
            column_name: column in table to use in determining partition.

        Returns:
            List of dicts of the form [{column_name: {"year": 2022}}]
        """
        return self.get_data_for_batch_identifiers_for_partition_on_date_parts(
            execution_engine=execution_engine,
            selectable=selectable,
            column_name=column_name,
            date_parts=[DatePart.YEAR],
        )

    def get_data_for_batch_identifiers_year_and_month(
        self,
        execution_engine: SqlAlchemyExecutionEngine,
        selectable: sqlalchemy.Selectable,
        column_name: str,
    ) -> List[dict]:
        """Build batch_identifiers from a column partition on year and month.

        This method builds a query to select the unique date_parts from the
        column_name. This data can be used to build BatchIdentifiers.

        Args:
            execution_engine: SqlAlchemyExecutionEngine to be used for executing the query.
            selectable: selectable to partition.
            column_name: column in table to use in determining partition.

        Returns:
            List of dicts of the form [{column_name: {"year": 2022, "month": 4}}]
        """
        return self.get_data_for_batch_identifiers_for_partition_on_date_parts(
            execution_engine=execution_engine,
            selectable=selectable,
            column_name=column_name,
            date_parts=[DatePart.YEAR, DatePart.MONTH],
        )

    def get_data_for_batch_identifiers_year_and_month_and_day(
        self,
        execution_engine: SqlAlchemyExecutionEngine,
        selectable: sqlalchemy.Selectable,
        column_name: str,
    ) -> List[dict]:
        """Build batch_identifiers from a column partition on year and month and day.

        This method builds a query to select the unique date_parts from the
        column_name. This data can be used to build BatchIdentifiers.

        Args:
            execution_engine: SqlAlchemyExecutionEngine to be used for executing the query.
            selectable: selectable to partition.
            column_name: column in table to use in determining partition.

        Returns:
            List of dicts of the form [{column_name: {"year": 2022, "month": 4, "day": 14}}]
        """
        return self.get_data_for_batch_identifiers_for_partition_on_date_parts(
            execution_engine=execution_engine,
            selectable=selectable,
            column_name=column_name,
            date_parts=[DatePart.YEAR, DatePart.MONTH, DatePart.DAY],
        )

    def get_partition_query_for_data_for_batch_identifiers_for_partition_on_date_parts(
        self,
        selectable: sqlalchemy.Selectable,
        column_name: str,
        date_parts: Union[List[DatePart], List[str]],
    ) -> sqlalchemy.Selectable:
        """Build a partition query to retrieve batch_identifiers info from a column partition on a list of date parts.

        This method builds a query to select the unique date_parts from the
        column_name. This data can be used to build BatchIdentifiers.

        Args:
            selectable: selectable to partition.
            column_name: column in table to use in determining partition.
            date_parts: part of the date to be used for partitioning e.g.
                DatePart.DAY or the case-insensitive string representation "day"

        Returns:
            List of dicts of the form [{column_name: {date_part_name: date_part_value}}]
        """  # noqa: E501 # FIXME CoP
        self._validate_date_parts(date_parts)

        date_parts = self._convert_date_parts(date_parts)

        # NOTE: AJB 20220414 concatenating to find distinct values to support all dialects.
        # There are more performant dialect-specific methods that can be implemented in
        # future improvements.
        # NOTE: AJB 20220511 for awsathena we need to cast extracted date parts
        # to string first before concatenating them.
        concat_clause: list[sqlalchemy.Label]
        concat_date_parts: sqlalchemy.Cast | sqlalchemy.ColumnOperators
        if len(date_parts) == 1:
            # MSSql does not accept single item concatenation
            concat_clause = sa.func.distinct(  # type: ignore[assignment] # FIXME CoP
                sa.func.extract(date_parts[0].value, sa.column(column_name)).label(
                    date_parts[0].value
                )
            ).label("concat_distinct_values")

        else:
            """
            # NOTE: <Alex>6/29/2022</Alex>
            Certain SQLAlchemy-compliant backends (e.g., Amazon Redshift, SQLite) allow only binary operators for "CONCAT".
            """  # noqa: E501 # FIXME CoP
            if self._dialect == GXSqlDialect.SQLITE:
                concat_date_parts = sa.cast(
                    sa.func.extract(date_parts[0].value, sa.column(column_name)),
                    sa.String,
                )

                date_part: DatePart
                for date_part in date_parts[1:]:
                    concat_date_parts = concat_date_parts.concat(
                        sa.cast(
                            sa.func.extract(date_part.value, sa.column(column_name)),
                            sa.String,
                        )
                    )

                concat_clause = sa.func.distinct(concat_date_parts).label("concat_distinct_values")  # type: ignore[assignment] # FIXME CoP
            else:
                concat_date_parts = sa.func.concat(
                    "",
                    sa.cast(
                        sa.func.extract(date_parts[0].value, sa.column(column_name)),
                        sa.String,
                    ),
                )

                for date_part in date_parts[1:]:
                    concat_date_parts = sa.func.concat(
                        concat_date_parts,
                        sa.cast(
                            sa.func.extract(date_part.value, sa.column(column_name)),
                            sa.String,
                        ),
                    )

                concat_clause = sa.func.distinct(concat_date_parts).label("concat_distinct_values")  # type: ignore[assignment] # FIXME CoP

        partitioned_query: sqlalchemy.Selectable = sa.select(  # type: ignore[call-overload] # FIXME CoP
            concat_clause,
            *[
                sa.cast(sa.func.extract(date_part.value, sa.column(column_name)), sa.Integer).label(
                    date_part.value
                )
                for date_part in date_parts
            ],
        ).select_from(selectable)

        return partitioned_query

    def get_data_for_batch_identifiers_for_partition_on_date_parts(
        self,
        execution_engine: SqlAlchemyExecutionEngine,
        selectable: sqlalchemy.Selectable,
        column_name: str,
        date_parts: Union[List[DatePart], List[str]],
    ) -> List[dict]:
        """Build batch_identifiers from a column partition on a list of date parts.

        This method builds a query to select the unique date_parts from the
        column_name. This data can be used to build BatchIdentifiers.

        Args:
            execution_engine: used to query the data to find batch identifiers.
            selectable: selectable to partition.
            column_name: column in table to use in determining partition.
            date_parts: part of the date to be used for partitioning e.g.
                DatePart.DAY or the case-insensitive string representation "day"

        Returns:
            List of dicts of the form [{column_name: {date_part_name: date_part_value}}]
        """

        partitioned_query: sqlalchemy.Selectable = (
            self.get_partition_query_for_data_for_batch_identifiers_for_partition_on_date_parts(
                selectable, column_name, date_parts
            )
        )

        result: List[sqlalchemy.Row | sqlalchemy.LegacyRow] = self._execute_partitioned_query(
            execution_engine, partitioned_query
        )

        return self._get_params_for_batch_identifiers_from_date_part_partitioner(
            column_name, result, date_parts
        )

    @staticmethod
    def _execute_partitioned_query(
        execution_engine: SqlAlchemyExecutionEngine,
        partitioned_query: sqlalchemy.Selectable,
    ) -> List[sqlalchemy.Row | sqlalchemy.LegacyRow]:
        """Use the provided execution engine to run the partition query and fetch all of the results.

        Args:
            execution_engine: SqlAlchemyExecutionEngine to be used for executing the query.
            partitioned_query: Query to be executed as a sqlalchemy Selectable.

        Returns:
            List of row results.
        """  # noqa: E501 # FIXME CoP
        return execution_engine.execute_partitioned_query(partitioned_query)

    def _get_params_for_batch_identifiers_from_date_part_partitioner(
        self,
        column_name: str,
        result: List[sqlalchemy.LegacyRow],
        date_parts: List[DatePart] | List[str],
    ) -> List[dict]:
        """Get parameters used to build BatchIdentifiers from the results of a get_data_for_batch_identifiers_for_partition_on_date_parts

        Args:
            column_name: Column name associated with the batch identifier.
            result: list of LegacyRow objects from sqlalchemy query result.
            date_parts: part of the date to be used for constructing the batch identifiers e.g.
                DatePart.DAY or the case-insensitive string representation "day"

        Returns:
            List of dicts of the form [{column_name: {date_part_name: date_part_value}}]
        """  # noqa: E501 # FIXME CoP
        date_parts = self._convert_date_parts(date_parts)

        data_for_batch_identifiers: List[dict] = [
            {
                column_name: {
                    date_part.value: getattr(row, date_part.value) for date_part in date_parts
                }
            }
            for row in result
        ]
        return data_for_batch_identifiers

    @staticmethod
    def _get_column_names_from_partitioner_kwargs(partitioner_kwargs) -> List[str]:
        column_names: List[str] = []

        if "column_names" in partitioner_kwargs:
            column_names = partitioner_kwargs["column_names"]
        elif "column_name" in partitioner_kwargs:
            column_names = [partitioner_kwargs["column_name"]]

        return column_names

    def get_data_for_batch_identifiers_for_non_date_part_partitioners(
        self,
        execution_engine: SqlAlchemyExecutionEngine,
        selectable: sqlalchemy.Selectable,
        partitioner_method_name: str,
        partitioner_kwargs: dict,
    ) -> List[dict]:
        """Build data used to construct batch identifiers for the input table using the provided partitioner config.

        Sql partitioner configurations yield the unique values that comprise a batch by introspecting your data.

        Args:
            execution_engine: Used to introspect the data.
            selectable: selectable to partition.
            partitioner_method_name: Desired partitioner method to use.
            partitioner_kwargs: Dict of directives used by the partitioner method as keyword arguments of key=value.

        Returns:
            List of dicts of the form [{column_name: {"key": value}}]
        """  # noqa: E501 # FIXME CoP
        get_partition_query_method_name: str = (
            self._get_method_name_for_get_data_for_batch_identifiers_method(partitioner_method_name)
        )

        partitioned_query: sqlalchemy.Selectable = getattr(self, get_partition_query_method_name)(
            selectable=selectable, **partitioner_kwargs
        )
        rows: List[sqlalchemy.LegacyRow] = self._execute_partitioned_query(
            execution_engine, partitioned_query
        )
        column_names: List[str] = self._get_column_names_from_partitioner_kwargs(partitioner_kwargs)
        return self._get_params_for_batch_identifiers_from_non_date_part_partitioners(
            column_names, rows
        )

    def _get_method_name_for_get_data_for_batch_identifiers_method(
        self, partitioner_method_name: str
    ):
        """Get the matching method name to get the data for batch identifiers from the input partitioner method name.

        Args:
            partitioner_method_name: Configured partitioner name.

        Returns:
            Name of the corresponding method to get data for building batch identifiers.
        """  # noqa: E501 # FIXME CoP
        processed_partitioner_method_name: str = self._get_partitioner_method_name(
            partitioner_method_name
        )
        try:
            return self.PARTITIONER_METHOD_TO_GET_UNIQUE_BATCH_IDENTIFIERS_METHOD_MAPPING[
                processed_partitioner_method_name
            ]
        except ValueError:
            raise gx_exceptions.InvalidConfigError(  # noqa: TRY003 # FIXME CoP
                f"Please provide a supported partitioner method name, you provided: {partitioner_method_name}"  # noqa: E501 # FIXME CoP
            )

    @staticmethod
    def _get_params_for_batch_identifiers_from_non_date_part_partitioners(
        column_names: List[str],
        rows: List[sqlalchemy.LegacyRow],
    ) -> List[dict]:
        """Get params used in batch identifiers from output of executing query for non date part partitioners.

        Args:
            column_names: Column names referenced in partitioner_kwargs.
            rows: Rows from execution of partition query.

        Returns:
            Dict of {column_name: row, column_name: row, ...}
        """  # noqa: E501 # FIXME CoP
        return [dict(zip(column_names, row)) for row in rows]

    @staticmethod
    def get_partition_query_for_data_for_batch_identifiers_for_partition_on_whole_table(
        selectable: sqlalchemy.Selectable,
    ) -> sqlalchemy.Selectable:
        """
        'Partition' by returning the whole table

        Note: the selectable parameter is a required to keep the signature of this method consistent with other methods.
        """  # noqa: E501 # FIXME CoP
        return sa.select(sa.true())

    @staticmethod
    def get_partition_query_for_data_for_batch_identifiers_for_partition_on_column_value(
        selectable: sqlalchemy.Selectable,
        column_name: str,
    ) -> sqlalchemy.Selectable:
        """Partition using the values in the named column"""
        return (
            sa.select(sa.func.distinct(sa.column(column_name)))
            .select_from(selectable)  # type: ignore[arg-type] # FIXME CoP
            .order_by(sa.column(column_name).asc())
        )

    def get_partition_query_for_data_for_batch_identifiers_for_partition_on_converted_datetime(
        self,
        selectable: sqlalchemy.Selectable,
        column_name: str,
        date_format_string: str = "%Y-%m-%d",
    ) -> sqlalchemy.Selectable:
        """Convert the values in the named column to the given date_format, and partition on that"""
        if self._dialect == "sqlite":
            return sa.select(
                sa.func.distinct(
                    sa.func.strftime(
                        date_format_string,
                        sa.column(column_name),
                    )
                )
            ).select_from(selectable)  # type: ignore[arg-type] # FIXME CoP

        raise NotImplementedError(
            f'Partitioner method "partition_on_converted_datetime" is not supported for "{self._dialect}" SQL dialect.'  # noqa: E501 # FIXME CoP
        )

    def get_partition_query_for_data_for_batch_identifiers_for_partition_on_divided_integer(
        self,
        selectable: sqlalchemy.Selectable,
        column_name: str,
        divisor: int,
    ) -> sqlalchemy.Selectable:
        """Divide the values in the named column by `divisor`, and partition on that"""
        if self._dialect == GXSqlDialect.SQLITE:
            return sa.select(
                sa.func.distinct(
                    sa.cast(
                        (sa.cast(sa.column(column_name), sa.Integer) / divisor),
                        sa.Integer,
                    )
                )
            ).select_from(selectable)  # type: ignore[arg-type] # FIXME CoP

        if self._dialect == GXSqlDialect.MYSQL:
            return sa.select(
                sa.func.distinct(
                    sa.cast(
                        sa.func.truncate(
                            (sa.cast(sa.column(column_name), sa.Integer) / divisor),
                            0,
                        ),
                        sa.Integer,
                    )
                )
            ).select_from(selectable)  # type: ignore[arg-type] # FIXME CoP

        if self._dialect == GXSqlDialect.MSSQL:
            return sa.select(
                sa.func.distinct(
                    sa.cast(
                        sa.func.round(
                            (sa.cast(sa.column(column_name), sa.Integer) / divisor),
                            0,
                            1,
                        ),
                        sa.Integer,
                    )
                )
            ).select_from(selectable)  # type: ignore[arg-type] # FIXME CoP

        if self._dialect == GXSqlDialect.AWSATHENA:
            return sa.select(
                sa.func.distinct(
                    sa.cast(
                        sa.func.truncate(sa.cast(sa.column(column_name), sa.Integer) / divisor),
                        sa.Integer,
                    )
                )
            ).select_from(selectable)  # type: ignore[arg-type] # FIXME CoP

        return sa.select(
            sa.func.distinct(
                sa.cast(
                    sa.func.trunc((sa.cast(sa.column(column_name), sa.Integer) / divisor), 0),
                    sa.Integer,
                )
            )
        ).select_from(selectable)  # type: ignore[arg-type] # FIXME CoP

    def get_partition_query_for_data_for_batch_identifiers_for_partition_on_mod_integer(
        self,
        selectable: sqlalchemy.Selectable,
        column_name: str,
        mod: int,
    ) -> sqlalchemy.Selectable:
        """Divide the values in the named column by `mod`, and partition on that"""
        if self._dialect in [
            GXSqlDialect.SQLITE,
            GXSqlDialect.MSSQL,
        ]:
            return sa.select(
                sa.func.distinct(sa.cast(sa.column(column_name), sa.Integer) % mod)
            ).select_from(selectable)  # type: ignore[arg-type] # FIXME CoP

        return sa.select(
            sa.func.distinct(sa.func.mod(sa.cast(sa.column(column_name), sa.Integer), mod))
        ).select_from(selectable)  # type: ignore[arg-type] # FIXME CoP

    @staticmethod
    def get_partition_query_for_data_for_batch_identifiers_for_partition_on_multi_column_values(
        selectable: sqlalchemy.Selectable,
        column_names: List[str],
    ) -> sqlalchemy.Selectable:
        """Partition on the joint values in the named columns"""
        return (
            sa.select(*[sa.column(column_name) for column_name in column_names])
            .distinct()
            .select_from(selectable)  # type: ignore[arg-type] # FIXME CoP
        )

    def get_partition_query_for_data_for_batch_identifiers_for_partition_on_hashed_column(
        self,
        selectable: sqlalchemy.Selectable,
        column_name: str,
        hash_digits: int,
    ) -> sqlalchemy.Selectable:
        """Note: this method is experimental. It does not work with all SQL dialects."""
        if self._dialect == GXSqlDialect.SQLITE:
            return sa.select(
                sa.func.distinct(
                    sa.func.md5(sa.cast(sa.column(column_name), sa.VARCHAR), hash_digits)
                )
            ).select_from(selectable)  # type: ignore[arg-type] # FIXME CoP

        raise NotImplementedError(
            f'Partitioner method "partition_on_hashed_column" is not supported for "{self._dialect}" SQL dialect.'  # noqa: E501 # FIXME CoP
        )



import pandas as pd

from great_expectations.datasource.fluent.interfaces import Batch
from great_expectations.metrics.column.values_not_match_regex_count import (
    ColumnValuesNotMatchRegexCount,
    ColumnValuesNotMatchRegexCountResult,
)
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.metrics.conftest import (
    ALL_DATA_SOURCES,
    SPARK_DATA_SOURCES,
    SQL_DATA_SOURCES,
    SnowflakeDatasourceTestConfig,
)

COLUMN_NAME = "whatevs"

DATA_FRAME = pd.DataFrame({COLUMN_NAME: ["abc", "def", "ghi", "1ab2", "1ab3", None]})

ALL_DATA_SOURCES_EXCEPT_SNOWFLAKE = [
    datasource
    for datasource in ALL_DATA_SOURCES
    if not isinstance(datasource, SnowflakeDatasourceTestConfig)
]


class TestColumnValuesNotMatchRegexCount:
    @parameterize_batch_for_data_sources(
        data_source_configs=ALL_DATA_SOURCES_EXCEPT_SNOWFLAKE,
        data=DATA_FRAME,
    )
    def test_partial_match_characters(self, batch_for_datasource: Batch) -> None:
        metric = ColumnValuesNotMatchRegexCount(column=COLUMN_NAME, regex="ab")
        metric_result = batch_for_datasource.compute_metrics(metric)

        assert isinstance(metric_result, ColumnValuesNotMatchRegexCountResult)
        assert metric_result.value == 2

    @parameterize_batch_for_data_sources(
        data_source_configs=SPARK_DATA_SOURCES + SQL_DATA_SOURCES,
        data=DATA_FRAME,
    )
    def test_special_characters(self, batch_for_datasource: Batch) -> None:
        metric = ColumnValuesNotMatchRegexCount(column=COLUMN_NAME, regex="^(a|d).+")
        metric_result = batch_for_datasource.compute_metrics(metric)

        assert isinstance(metric_result, ColumnValuesNotMatchRegexCountResult)
        assert metric_result.value == 3



from __future__ import annotations

from copy import deepcopy
from typing import TYPE_CHECKING, Any, ClassVar, Dict, Optional, Type, Union

from great_expectations.compatibility import pydantic
from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.suite_parameters import (
    SuiteParameterDict,  # noqa: TC001 # FIXME CoP
)
from great_expectations.expectations.expectation import (
    BatchExpectation,
    render_suite_parameter_string,
)
from great_expectations.expectations.metadata_types import DataQualityIssues, SupportedDataSources
from great_expectations.expectations.model_field_types import (
    ConditionParser,  # noqa: TC001 # FIXME CoP
)
from great_expectations.render import (
    LegacyDiagnosticRendererType,
    LegacyRendererType,
    RenderedStringTemplateContent,
)
from great_expectations.render.renderer.renderer import renderer
from great_expectations.render.renderer_configuration import (
    RendererConfiguration,
    RendererValueType,
)
from great_expectations.render.util import num_to_str, substitute_none_for_missing
from great_expectations.validator.metric_configuration import (
    MetricConfiguration,
)

if TYPE_CHECKING:
    from great_expectations.core import (
        ExpectationValidationResult,
    )
    from great_expectations.execution_engine import ExecutionEngine
    from great_expectations.expectations.expectation_configuration import (
        ExpectationConfiguration,
    )
    from great_expectations.validator.validator import ValidationDependencies

EXPECTATION_SHORT_DESCRIPTION = (
    "Expect the number of rows to equal the number in another table within the same database."
)
OTHER_TABLE_NAME_DESCRIPTION = (
    "The name of the other table. Other table must be located within the same database."
)
SUPPORTED_DATA_SOURCES = [
    SupportedDataSources.SQLITE.value,
    SupportedDataSources.POSTGRESQL.value,
    SupportedDataSources.REDSHIFT.value,
    SupportedDataSources.MYSQL.value,
    SupportedDataSources.MSSQL.value,
    SupportedDataSources.DATABRICKS.value,
    SupportedDataSources.SNOWFLAKE.value,
]
DATA_QUALITY_ISSUES = [DataQualityIssues.VOLUME.value]


class ExpectTableRowCountToEqualOtherTable(BatchExpectation):
    __doc__ = f"""{EXPECTATION_SHORT_DESCRIPTION}

    ExpectTableRowCountToEqualOtherTable is a \
    Batch Expectation.

    BatchExpectations are one of the most common types of Expectation.
    They are evaluated for an entire Batch, and answer a semantic question about the Batch itself.

    Args:
        other_table_name (str): {OTHER_TABLE_NAME_DESCRIPTION}

    Other Parameters:
        result_format (str or None): \
            Which output mode to use: BOOLEAN_ONLY, BASIC, COMPLETE, or SUMMARY. \
            For more detail, see [result_format](https://docs.greatexpectations.io/docs/reference/expectations/result_format).
        catch_exceptions (boolean or None): \
            If True, then catch exceptions and include them as part of the result object. \
            For more detail, see [catch_exceptions](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#catch_exceptions).
        meta (dict or None): \
            A JSON-serializable dictionary (nesting allowed) that will be included in the output without \
            modification. For more detail, see [meta](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#meta).

    Returns:
        An [ExpectationSuiteValidationResult](https://docs.greatexpectations.io/docs/terms/validation_result)

        Exact fields vary depending on the values passed to result_format, catch_exceptions, and meta.

    See Also:
        [ExpectTableRowCountToBeBetween](https://greatexpectations.io/expectations/expect_table_row_count_to_be_between)
        [ExpectTableRowCountToEqual](https://greatexpectations.io/expectations/expect_table_row_count_to_equal)

    Supported Data Sources:
        [{SUPPORTED_DATA_SOURCES[0]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[1]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[2]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[3]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[4]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[5]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[6]}](https://docs.greatexpectations.io/docs/application_integration_support/)

    Data Quality Issues:
        {DATA_QUALITY_ISSUES[0]}

    Example Data:
            test_table
                test 	test2
            0 	1.00 	2
            1 	2.30 	5
            2 	4.33 	0

            test_table_two
                test 	test2
            0 	1.00 	2
            1 	2.30 	5
            2 	4.33 	0

            test_table_three
                test 	test2
            0 	1.00 	2
            1 	2.30 	5

    Code Examples:
        Passing Case:
            Input:
                ExpectTableRowCountToEqualOtherTable(
                    other_table_name=test_table_two
            )

            Output:
                {{
                  "exception_info": {{
                    "raised_exception": false,
                    "exception_traceback": null,
                    "exception_message": null
                  }},
                  "result": {{
                    "observed_value": 3
                  }},
                  "meta": {{}},
                  "success": true
                }}

        Failing Case:
            Input:
                ExpectTableRowCountToEqualOtherTable(
                    other_table_name=test_table_three
            )

            Output:
                {{
                  "exception_info": {{
                    "raised_exception": false,
                    "exception_traceback": null,
                    "exception_message": null
                  }},
                  "result": {{
                    "observed_value": 2
                  }},
                  "meta": {{}},
                  "success": false
                }}
    """  # noqa: E501 # FIXME CoP

    other_table_name: Union[str, SuiteParameterDict] = pydantic.Field(
        description=OTHER_TABLE_NAME_DESCRIPTION
    )
    row_condition: Union[str, None] = None
    condition_parser: Union[ConditionParser, None] = None

    library_metadata: ClassVar[Dict[str, Union[str, list, bool]]] = {
        "maturity": "production",
        "tags": ["core expectation", "table expectation", "multi-table expectation"],
        "contributors": [
            "@great_expectations",
        ],
        "requirements": [],
        "has_full_test_suite": True,
        "manually_reviewed_code": True,
    }
    _library_metadata = library_metadata

    metric_dependencies = ("table.row_count",)
    success_keys = ("other_table_name",)
    args_keys = ("other_table_name",)

    class Config:
        title = "Expect table row count to equal other table"

        @staticmethod
        def schema_extra(
            schema: Dict[str, Any], model: Type[ExpectTableRowCountToEqualOtherTable]
        ) -> None:
            BatchExpectation.Config.schema_extra(schema, model)
            schema["properties"]["metadata"]["properties"].update(
                {
                    "data_quality_issues": {
                        "title": "Data Quality Issues",
                        "type": "array",
                        "const": DATA_QUALITY_ISSUES,
                    },
                    "library_metadata": {
                        "title": "Library Metadata",
                        "type": "object",
                        "const": model._library_metadata,
                    },
                    "short_description": {
                        "title": "Short Description",
                        "type": "string",
                        "const": EXPECTATION_SHORT_DESCRIPTION,
                    },
                    "supported_data_sources": {
                        "title": "Supported Data Sources",
                        "type": "array",
                        "const": SUPPORTED_DATA_SOURCES,
                    },
                }
            )

    @override
    @classmethod
    def _prescriptive_template(
        cls,
        renderer_configuration: RendererConfiguration,
    ) -> RendererConfiguration:
        renderer_configuration.add_param(
            name="other_table_name", param_type=RendererValueType.STRING
        )
        renderer_configuration.template_str = (
            "Row count must equal the row count of table $other_table_name."
        )
        return renderer_configuration

    @override
    @classmethod
    @renderer(renderer_type=LegacyRendererType.PRESCRIPTIVE)
    @render_suite_parameter_string
    def _prescriptive_renderer(
        cls,
        configuration: Optional[ExpectationConfiguration] = None,
        result: Optional[ExpectationValidationResult] = None,
        runtime_configuration: Optional[dict] = None,
        **kwargs,
    ):
        runtime_configuration = runtime_configuration or {}
        styling = runtime_configuration.get("styling")
        if not configuration:
            raise ValueError("configuration is required for prescriptive renderer")  # noqa: TRY003 # FIXME CoP
        params = substitute_none_for_missing(configuration.kwargs, ["other_table_name"])
        template_str = "Row count must equal the row count of table $other_table_name."

        return [
            RenderedStringTemplateContent(
                content_block_type="string_template",
                string_template={
                    "template": template_str,
                    "params": params,
                    "styling": styling,
                },
            )
        ]

    @override
    @classmethod
    @renderer(renderer_type=LegacyDiagnosticRendererType.OBSERVED_VALUE)
    def _diagnostic_observed_value_renderer(
        cls,
        configuration: Optional[ExpectationConfiguration] = None,
        result: Optional[ExpectationValidationResult] = None,
        runtime_configuration: Optional[dict] = None,
        **kwargs,
    ):
        if not result or result.result.get("observed_value"):
            return "--"

        self_table_row_count = num_to_str(result.result["observed_value"]["self"])
        other_table_row_count = num_to_str(result.result["observed_value"]["other"])

        return RenderedStringTemplateContent(
            content_block_type="string_template",
            string_template={
                "template": "Row Count: $self_table_row_count<br>Other Table Row Count: $other_table_row_count",  # noqa: E501 # FIXME CoP
                "params": {
                    "self_table_row_count": self_table_row_count,
                    "other_table_row_count": other_table_row_count,
                },
                "styling": {"classes": ["mb-2"]},
            },
        )

    @override
    def get_validation_dependencies(
        self,
        execution_engine: Optional[ExecutionEngine] = None,
        runtime_configuration: Optional[dict] = None,
    ) -> ValidationDependencies:
        validation_dependencies: ValidationDependencies = super().get_validation_dependencies(
            execution_engine, runtime_configuration
        )

        configuration = self.configuration
        kwargs = configuration.kwargs if configuration else {}
        other_table_name = kwargs.get("other_table_name")

        # At this time, this is the only Expectation that
        # computes the same metric over more than one domain
        # ValidationDependencies does not allow duplicate metric names
        # and the registry is checked to ensure the metric name is registered
        # as a side effect of the super().get_validation_dependencies() call above
        # As a work-around, after the registry check
        # we create a second table.row_count metric for the other table manually
        # and rename the metrics defined in ValidationDependencies
        table_row_count_metric_config_self: Optional[MetricConfiguration] = (
            validation_dependencies.get_metric_configuration(metric_name="table.row_count")
        )
        assert table_row_count_metric_config_self, "table_row_count_metric should not be None"
        copy_table_row_count_metric_config_self = deepcopy(table_row_count_metric_config_self)
        copy_table_row_count_metric_config_self.metric_domain_kwargs["table"] = other_table_name
        # instantiating a new MetricConfiguration gives us a new id
        table_row_count_metric_config_other = MetricConfiguration(
            metric_name="table.row_count",
            metric_domain_kwargs=copy_table_row_count_metric_config_self.metric_domain_kwargs,
            metric_value_kwargs=copy_table_row_count_metric_config_self.metric_value_kwargs,
        )
        # rename original "table.row_count" metric to "table.row_count.self"
        validation_dependencies.set_metric_configuration(
            metric_name="table.row_count.self",
            metric_configuration=table_row_count_metric_config_self,
        )
        validation_dependencies.remove_metric_configuration(metric_name="table.row_count")
        # add a new metric dependency named "table.row_count.other" with modified metric config
        validation_dependencies.set_metric_configuration(
            "table.row_count.other", table_row_count_metric_config_other
        )
        return validation_dependencies

    @override
    def _validate(
        self,
        metrics: Dict,
        runtime_configuration: Optional[dict] = None,
        execution_engine: Optional[ExecutionEngine] = None,
    ):
        table_row_count_self = metrics["table.row_count.self"]
        table_row_count_other = metrics["table.row_count.other"]

        return {
            "success": table_row_count_self == table_row_count_other,
            "result": {
                "observed_value": {
                    "self": table_row_count_self,
                    "other": table_row_count_other,
                }
            },
        }
